#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser."""

from ramses_rf import Device
from ramses_rf.device import Fakeable as _Fakeable


def ensure_fakeable(dev: Device) -> None:
    """If a Device is not Fakeable, make it so."""

    class Fakeable(dev.__class__, _Fakeable):
        pass

    if isinstance(dev, _Fakeable):
        # if hasattr(dev, "_make_fake"):  # no need for callable(getattr(...))
        return

    dev.__class__ = Fakeable
    setattr(dev, "_faked", None)
    setattr(dev, "_context", None)
    setattr(dev, "_1fc9_state", {})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

# TODO: test addenda phase of binding handshake
# TODO: get test working with (and without) disabled QoS

"""Test the binding protocol with a virtual RF

    NB: This test will likely fail with pytest-repeat (pytest -n x); maybe because of
    concurrent access to pty.openpty().
"""

import asyncio
from datetime import datetime as dt
from unittest.mock import patch

import pytest

from ramses_rf import Code, Device, Gateway, Message, Packet
from ramses_rf.binding_fsm import (
    SZ_RESPONDENT,
    SZ_SUPPLICANT,
    BindContext,
    BindStateBase,
    _BindStates,
)
from ramses_rf.device import Fakeable

from .virtual_rf import rf_factory

# patched constants
_DEBUG_DISABLE_IMPERSONATION_ALERTS = True  # # ramses_tx.protocol
_DEBUG_DISABLE_QOS = False  # #                 ramses_tx.protocol
DEFAULT_MAX_RETRIES = 0  # #                    ramses_tx.protocol
DEFAULT_TIMEOUT = 0.005  # #                    ramses_tx.protocol_fsm
MAINTAIN_STATE_CHAIN = False  # #               ramses_tx.protocol_fsm
MIN_GAP_BETWEEN_WRITES = 0  # #                 ramses_tx.protocol

# other constants
ASSERT_CYCLE_TIME = 0.0005  # max_cycles_per_assert = max_sleep / ASSERT_CYCLE_TIME
DEFAULT_MAX_SLEEP = 0.1

PKT_FLOW = "packets"


DEFAULT_GATEWAY_CONFIG = {
    "config": {
        "disable_discovery": True,
        "enforce_known_list": True,
    }
}

ITHO__ = "itho"
NUAIRE = "nuaire"
ORCON_ = "orcon"

TEST_SUITE_300 = [
    {
        SZ_RESPONDENT: {
            "30:098165": {"class": "FAN", " scheme": "nuaire", "_notes": "ECO-HEAT-HC"},
        },
        SZ_SUPPLICANT: {
            "32:208628": {"class": "REM", "scheme": "nuaire", "_notes": "4-way switch"}
        },
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._31DA], "idx": "21"},
        PKT_FLOW: (
            " I --- 32:208628 --:------ 32:208628 1FC9 018 00-22F1-832EF4 6C-10E0-832EF4 00-1FC9-832EF4",
            " W --- 30:098165 32:208628 --:------ 1FC9 006 21-31DA-797F75",
            " I --- 32:208628 30:098165 --:------ 1FC9 001 21",
            " I --- 32:208628 63:262142 --:------ 10E0 030 00-0001C85A0101-6C-FFFFFFFFFFFF010607E0564D4E2D32334C4D48323300",
            # I --- 32:208628 --:------ 32:208628 1060 003 00-FF01",  # sends x3
        ),
    },
    {
        SZ_RESPONDENT: {
            "18:126620": {"class": "FAN", "scheme": "itho", "_notes": "Spider HRU"},
        },
        SZ_SUPPLICANT: {
            "37:154011": {"class": "CO2", "scheme": "itho", "_notes": "Spider CO2"}
        },
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._31D9, Code._31DA]},
        PKT_FLOW: (
            " I --- 37:154011 --:------ 37:154011 1FC9 030 00-31E0-96599B 00-1298-96599B 00-2E10-96599B 01-10E0-96599B 00-1FC9-96599B",
            " W --- 18:126620 37:154011 --:------ 1FC9 012 00-31D9-49EE9C 00-31DA-49EE9C",
            " I --- 37:154011 18:126620 --:------ 1FC9 001 00",
            " I --- 37:154011 63:262142 --:------ 10E0 038 00-000100280901-01-FEFFFFFFFFFF140107E5564D532D31324333390000000000000000000000",
        ),
    },
    {  # FIXME: offer sent to 63:262142, so send_cmd() wont return corresponding accept
        SZ_RESPONDENT: {
            "32:155617": {"class": "FAN", "scheme": "orcon", "_notes": "HRC-350"},
        },
        SZ_SUPPLICANT: {
            "29:158183": {"class": "REM", "scheme": "orcon", "_notes": "VMN-15LF01"}
        },
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._31D9, Code._31DA]},
        PKT_FLOW: (
            " I --- 29:158183 63:262142 --:------ 1FC9 024 00-22F1-7669E7 00-22F3-7669E7 67-10E0-7669E7 00-1FC9-7669E7",
            " W --- 32:155617 29:158183 --:------ 1FC9 012 00-31D9-825FE1 00-31DA-825FE1",
            " I --- 29:158183 32:155617 --:------ 1FC9 001 00",
            " I --- 29:158183 63:262142 --:------ 10E0 038 00-0001C8270901-67-FFFFFFFFFFFF0D0207E3564D4E2D31354C46303100000000000000000000",
            # I --- 29:158183 --:------ 29:158183 1060 003 00-FF01",
        ),
    },
    {  # FIXME: supplicant used oem_code and 10E0
        SZ_RESPONDENT: {"32:155617": {"class": "FAN", "scheme": "orcon"}},
        SZ_SUPPLICANT: {"37:171871": {"class": "DIS", "scheme": "orcon"}},
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._31D9, Code._31DA]},
        PKT_FLOW: (
            " I --- 37:171871 --:------ 37:171871 1FC9 024 00-22F1-969F5F 00-22F3-969F5F 67-10E0-969F5F 00-1FC9-969F5F",
            " W --- 32:155617 37:171871 --:------ 1FC9 012 00-31D9-825FE1 00-31DA-825FE1",
            " I --- 37:171871 32:155617 --:------ 1FC9 001 00",
            " I --- 37:171871 63:262142 --:------ 10E0 038 00-0001C8940301-67-FFFFFFFFFFFF1B0807E4564D492D313557534A3533000000000000000000",
        ),
    },
    {  # FIXME: confirm is:  I --- 07:045960 01:145038 --:------ 1FC9 006 0012601CB388
        SZ_RESPONDENT: {"01:145038": {"class": "CTL"}},
        SZ_SUPPLICANT: {"07:045960": {"class": "DHW"}},
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._10A0]},
        PKT_FLOW: (
            " I --- 07:045960 --:------ 07:045960 1FC9 012 00-1260-1CB388 00-1FC9-1CB388",
            " W --- 01:145038 07:045960 --:------ 1FC9 006 00-10A0-06368E",
            " I --- 07:045960 01:145038 --:------ 1FC9 006 00-1260-1CB388",
        ),  # TODO: need epilogue packets, if any (1060?)
    },
    {  # FIXME: confirm is not:  I --- 22:057520 01:085545 --:------ 1FC9 006 07
        SZ_RESPONDENT: {"01:085545": {"class": "CTL"}},
        SZ_SUPPLICANT: {"22:057520": {"class": "THM"}},  # is THM, not STA
        f"{SZ_RESPONDENT}_attr": {"codes": [Code._2309], "idx": "07"},
        PKT_FLOW: (
            " I --- 22:057520 --:------ 22:057520 1FC9 024 00-2309-58E0B0 00-30C9-58E0B0 00-0008-58E0B0 00-1FC9-58E0B0",
            " W --- 01:085545 22:057520 --:------ 1FC9 006 07-2309-054E29",
            " I --- 22:057520 01:085545 --:------ 1FC9 006 00-2309-58E0B0",
        ),
    },
    {  # FIXME: needs initiate_binding_process(), and above
        SZ_RESPONDENT: {"01:145038": {"class": "CTL"}},
        SZ_SUPPLICANT: {"34:092243": {"class": "RND"}},
        PKT_FLOW: (
            " I --- 34:259472 --:------ 34:259472 1FC9 024 00-2309-8BF590 00-30C9-8BF590 00-0008-8BF590 00-1FC9-8BF590",
            " W --- 01:220768 34:259472 --:------ 1FC9 006 01-2309-075E60",
            " I --- 34:259472 01:220768 --:------ 1FC9 006 01-2309-8BF590",
            " I --- 34:259472 63:262142 --:------ 10E0 038 00-0001C8380F01-00-F1FF070B07E6030507E15438375246323032350000000000000000000000",
            # I --- 34:259472 --:------ 34:259472 1060 003 00-FF01",
            # I --- 34:259472 --:------ 34:259472 0005 012 000A0000000F000000100000",
            # I --- 34:259472 --:------ 34:259472 000C 018 000A7FFFFFFF000F7FFFFFFF00107FFFFFFF",
        ),
    },
]

RESPONDENT_ATTRS_BY_SUPPLICANT = {
    f"{list(d[SZ_SUPPLICANT].keys())[0]} to {list(d[SZ_RESPONDENT].keys())[0]}": d.get(
        f"{SZ_RESPONDENT}_attr", {}
    )  # can't use .pop in a comprehension
    for d in TEST_SUITE_300
}

TEST_SUITE_300 = [
    {k: d[k] for k in (SZ_SUPPLICANT, SZ_RESPONDENT, PKT_FLOW)} for d in TEST_SUITE_300
]


# ### FIXTURES #########################################################################


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(
        "ramses_tx.protocol._DEBUG_DISABLE_IMPERSONATION_ALERTS",
        _DEBUG_DISABLE_IMPERSONATION_ALERTS,
    )
    monkeypatch.setattr(
        "ramses_tx.protocol.MIN_GAP_BETWEEN_WRITES", MIN_GAP_BETWEEN_WRITES
    )


def pytest_generate_tests(metafunc: pytest.Metafunc):
    def id_fnc(test_set):
        r_class = list(test_set[SZ_RESPONDENT].values())[0]["class"]
        s_class = list(test_set[SZ_SUPPLICANT].values())[0]["class"]
        return s_class + " binding to " + r_class

    metafunc.parametrize("test_set", TEST_SUITE_300, ids=id_fnc)


# ######################################################################################


def ensure_fakeable(dev: Device) -> None:
    """If a Device is not Fakeable (i.e. Fakeable, not _faked), make it so."""

    class _Fakeable(dev.__class__, Fakeable):
        pass

    if isinstance(dev, Fakeable):
        # if hasattr(dev, "_make_fake"):  # no need for callable(getattr(...))
        return

    dev.__class__ = _Fakeable
    setattr(dev, "_faked", None)
    setattr(dev, "_context", BindContext(dev))
    setattr(dev, "_1fc9_state", {})


async def assert_context_state(
    device: Fakeable, state: type[BindStateBase], max_sleep: int = DEFAULT_MAX_SLEEP
) -> None:
    for _ in range(int(max_sleep / ASSERT_CYCLE_TIME)):
        await asyncio.sleep(ASSERT_CYCLE_TIME)
        if isinstance(device._context.state, state):
            break
    assert isinstance(device._context.state, state)


# ### TESTS ############################################################################


# TODO: test addenda phase of binding handshake
async def _test_flow_10x(
    gwy_r: Gateway, gwy_s: Gateway, pkt_flow_expected: list[str]
) -> None:
    """Check the change of state during a binding at device layer."""

    # STEP 0: Setup...
    loop = asyncio.get_running_loop()

    respondent: Fakeable = gwy_r.devices[0]
    supplicant: Fakeable = gwy_s.devices[0]
    ensure_fakeable(respondent)

    await assert_context_state(respondent, _BindStates.IS_IDLE_DEVICE)
    await assert_context_state(supplicant, _BindStates.IS_IDLE_DEVICE)

    assert not respondent._context.is_binding
    assert not supplicant._context.is_binding

    #
    # Step R0: Respondent initial state
    respondent._context.set_state(_BindStates.NEEDING_TENDER)
    await assert_context_state(respondent, _BindStates.NEEDING_TENDER)
    assert respondent._context.is_binding

    #
    # Step S0: Supplicant initial state
    supplicant._context.set_state(_BindStates.NEEDING_ACCEPT)
    await assert_context_state(supplicant, _BindStates.NEEDING_ACCEPT)
    assert supplicant._context.is_binding

    #
    # Step R1: Respondent expects an Offer
    resp_task = loop.create_task(respondent._context._wait_for_offer())

    #
    # Step S1: Supplicant sends an Offer (makes Offer) and expects an Accept
    msg = Message(Packet(dt.now(), "000 " + pkt_flow_expected[0]))
    codes = [b[1] for b in msg.payload["bindings"] if b[1] != Code._1FC9]

    pkt = await supplicant._context._make_offer(codes)
    await assert_context_state(supplicant, _BindStates.NEEDING_ACCEPT)

    await resp_task
    await assert_context_state(respondent, _BindStates.NEEDING_AFFIRM)

    tender = resp_task.result()
    assert tender._pkt == pkt, "Resp's Msg doesn't match Supp's Offer cmd"

    supp_task = loop.create_task(supplicant._context._wait_for_accept(tender))

    #
    # Step R2: Respondent expects a Confirm after sending an Accept (accepts Offer)
    msg = Message(Packet(dt.now(), "000 " + pkt_flow_expected[1]))
    codes = [b[1] for b in msg.payload["bindings"]]

    pkt = await respondent._context._accept_offer(tender, codes)
    await assert_context_state(respondent, _BindStates.NEEDING_AFFIRM)

    await supp_task
    await assert_context_state(supplicant, _BindStates.TO_SEND_AFFIRM)

    accept = supp_task.result()
    assert accept._pkt == pkt, "Supp's Msg doesn't match Resp's Accept cmd"

    resp_task = loop.create_task(respondent._context._wait_for_confirm(accept))

    #
    # Step S2: Supplicant sends a Confirm (confirms Accept)
    msg = Message(Packet(dt.now(), "000 " + pkt_flow_expected[2]))
    codes = [b[1] for b in msg.payload["bindings"] if len(b) > 1]

    pkt = await supplicant._context._confirm_accept(accept, codes=codes)
    if True or len(pkt_flow_expected) == 3:  # FIXME
        await assert_context_state(supplicant, _BindStates.HAS_BOUND_SUPP)
    else:
        await assert_context_state(supplicant, _BindStates.TO_SEND_RATIFY)

    await resp_task
    if True or len(pkt_flow_expected) == 3:  # FIXME
        await assert_context_state(respondent, _BindStates.HAS_BOUND_RESP)
    else:
        await assert_context_state(respondent, _BindStates.NEEDING_RATIFY)

    affirm = resp_task.result()
    assert affirm._pkt == pkt, "Resp's Msg doesn't match Supp's Confirm cmd"

    #
    # Some bindings don't include an Addenda...
    if True or len(pkt_flow_expected) == 3:  # i.e. no addenda  FIXME
        return

    await assert_context_state(respondent, _BindStates.NEEDING_RATIFY)
    await assert_context_state(supplicant, _BindStates.TO_SEND_RATIFY)

    # Step R3: Respondent expects an Addenda (optional)
    resp_task = loop.create_task(
        respondent._context._wait_for_addenda(accept, timeout=0.05)
    )

    # Step S3: Supplicant sends an Addenda (optional)
    msg = Message(Packet(dt.now(), "000 " + pkt_flow_expected[3]))
    supplicant._msgz[msg.code] = {msg.verb: {msg._pkt._ctx: msg}}

    pkt = await supplicant._context._cast_addenda()
    await assert_context_state(supplicant, _BindStates.HAS_BOUND_SUPP)

    await assert_context_state(respondent, _BindStates.HAS_BOUND_RESP)
    await resp_task

    ratify = resp_task.result()
    assert ratify._pkt == pkt, "Resp's Msg doesn't match Supp's Addenda cmd"

    assert False


# TODO: get test working without QoS
@pytest.mark.xdist_group(name="virt_serial")
@patch("ramses_tx.protocol._DEBUG_DISABLE_QOS", _DEBUG_DISABLE_QOS)
async def test_flow_100(test_set: dict[str:dict]) -> None:
    """Check packet flow / state change of a binding at device layer."""

    config = {}
    for role in (SZ_RESPONDENT, SZ_SUPPLICANT):
        devices = [d for d in test_set.values() if isinstance(d, dict)]
        config[role] = DEFAULT_GATEWAY_CONFIG | {
            "known_list": {k: v for d in devices for k, v in d.items()},
            "orphans_hvac": list(test_set[role]),  # TODO: used by Heat domain too!
        }

    pkt_flow = [
        x[:46] + x[46:].replace(" ", "").replace("-", "")
        for x in test_set.get(PKT_FLOW, [])
    ]

    # cant use fixture for this, as new schema required for every test
    rf, gwys = await rf_factory([config[SZ_RESPONDENT], config[SZ_SUPPLICANT]])

    try:
        await _test_flow_10x(gwys[0], gwys[1], pkt_flow)
    finally:
        for gwy in gwys:
            await gwy.stop()
        await rf.stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the Schema processor.
"""

from pathlib import Path, PurePath

from ramses_rf import Gateway
from ramses_rf.helpers import shrink
from tests.helpers import TEST_DIR  # noqa: F401

WORK_DIR = f"{TEST_DIR}/logs"

SCHEMA_EMPTY = {"known_list": {}, "main_tcs": None, "orphans": []}


def id_fnc(param):
    return PurePath(param).name


def pytest_generate_tests(metafunc):
    metafunc.parametrize("f_name", Path(WORK_DIR).glob("*.log"), ids=id_fnc)


async def test_log_file(f_name):
    with open(f_name) as f:
        gwy = Gateway(None, input_file=f, config={})
        await gwy.start()

    assert shrink(gwy.schema) == shrink(SCHEMA_EMPTY)

    await gwy.stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the various helper APIs.
"""

# TODO: add test for ramses_tx.frame.pkt_header()

from ramses_rf.system.zones import _transform
from ramses_tx.helpers import (
    hex_from_bool,
    hex_from_double,
    hex_from_dtm,
    hex_from_dts,
    hex_from_flag8,
    hex_from_temp,
    hex_to_bool,
    hex_to_double,
    hex_to_dtm,
    hex_to_dts,
    hex_to_flag8,
    hex_to_temp,
)
from ramses_tx.packet import Packet
from tests.helpers import TEST_DIR

WORK_DIR = f"{TEST_DIR}/parser_helpers"


def test_pkt_addr_parser():  # noqa: F811
    def proc_log_line(pkt_line):
        if "#" not in pkt_line:
            return

        pkt_line, pkt_dict = pkt_line.split("#", maxsplit=1)

        if not pkt_line[27:].strip():
            return

        pkt = Packet.from_file(pkt_line[:26], pkt_line[27:])

        assert (pkt.src.id, pkt.dst.id) == eval(pkt_dict)

    with open(f"{WORK_DIR}/pkt_addrs.log") as f:
        while line := (f.readline()):
            if line.strip():
                proc_log_line(line)


def test_demand_transform() -> None:
    assert [x[1] for x in TRANSFORMS] == [_transform(x[0]) for x in TRANSFORMS]


def test_field_parsers() -> None:
    for val in ("FF", "00", "C8"):
        assert val == hex_from_bool(hex_to_bool(val))

    for val in ("7FFF", "0000", "0001", "0010", "0100", "1000"):
        assert val == hex_from_double(hex_to_double(val))
        assert val == hex_from_double(hex_to_double(val, factor=100), factor=100)

    for val in (
        "FF" * 6,
        "FF" * 7,
        "00141B0A07E3",
        "00110E0507E5",
        "0400041C0A07E3",
    ):
        assert val == hex_from_dtm(hex_to_dtm(val), incl_seconds=(len(val) == 14))

    for val in ("00000000007F",):
        assert val == hex_from_dts(hex_to_dts(val))

    for val in ("00", "01", "08", "10", "E0", "CC", "FF"):
        assert val == hex_from_flag8(hex_to_flag8(val))
        assert val == hex_from_flag8(hex_to_flag8(val, lsb=True), lsb=True)

    for val in ("7FFF", "7EFF", "0000", "0010", "0200", "D000"):
        assert val == hex_from_temp(hex_to_temp(val))

    for val in (None, False, -127.99, -100, -22.5, -1.53, 0, 1.53, 22.5, 100, 127.98):
        assert val == hex_to_temp(hex_from_temp(val))


TRANSFORMS = [
    (0.000, 0),
    (0.220, 0),
    (0.230, 0),
    (0.260, 0),
    # (0.295, 0),  # needs confirming
    (0.300, 0),
    # (0.305, 0),  # needs confirming
    (0.310, 0.01),
    (0.340, 0.03),
    (0.350, 0.04),
    (0.370, 0.05),
    (0.380, 0.06),
    (0.390, 0.07),
    (0.400, 0.08),
    (0.410, 0.08),
    (0.420, 0.09),
    (0.430, 0.10),
    (0.450, 0.11),
    (0.470, 0.13),
    (0.480, 0.14),
    (0.530, 0.17),
    (0.540, 0.18),
    (0.550, 0.19),
    (0.560, 0.20),
    (0.575, 0.21),
    (0.610, 0.23),
    (0.620, 0.24),
    (0.650, 0.26),
    (0.660, 0.27),
    (0.680, 0.29),
    (0.690, 0.29),
    # (0.695, 0.30),  # needs confirming
    (0.700, 0.30),
    # (0.705, 0.31),  # needs confirming
    (0.720, 0.35),
    (0.740, 0.39),
    (0.760, 0.44),
    (0.770, 0.46),
    (0.790, 0.51),
    (0.800, 0.53),
    (0.820, 0.58),
    (0.830, 0.60),
    (0.840, 0.63),
    (0.930, 0.84),
    (0.950, 0.88),
    # (0.995, 0.99),  # needs confirming
    (1.000, 1.0),
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test binding of Heat devices.
"""

import asyncio
from datetime import datetime as dt
from datetime import timedelta as td

from tests_deprecated.common import (
    MOCKED_PORT,
    MockGateway,
    abort_if_rf_test_fails,
    load_test_gwy,
)

from ramses_rf.binding_fsm import _BindStates
from ramses_rf.const import Code, DevType
from ramses_rf.device import HvacRemote, HvacVentilator
from ramses_rf.device.base import Fakeable
from ramses_rf.schemas import SZ_CONFIG, SZ_DISABLE_DISCOVERY, SZ_ORPHANS_HVAC
from ramses_tx import Message
from ramses_tx.schemas import SZ_CLASS, SZ_ENFORCE_KNOWN_LIST, SZ_FAKED, SZ_KNOWN_LIST

FAN_ID = "21:111111"
REM_ID = "33:333333"
CO2_ID = "44:444444"


class HvacVentilatorFakable(HvacVentilator, Fakeable):
    pass


def pytest_generate_tests(metafunc):
    test_ports = {MOCKED_PORT: MockGateway}

    metafunc.parametrize("test_port", test_ports.items(), ids=test_ports.keys())


BIND_REQUEST_EXPECTED = 10
BIND_OFFER_EXPECTED = 11
BIND_CONFIRM_EXPECTED = 12
BIND_COMPLETED = 13

# NOTE: used as a global
flow_marker: int = None  # type: ignore[assignment]


def track_packet_flow(msg, requestor_id, listener_id, *args, **kwargs):
    """Test the flow of packets (messages) as initiated by the requestor."""

    global flow_marker

    if msg.code not in (Code._1FC9,):
        return

    # track the 3-way handshake
    if msg._pkt._hdr == "1FC9| I|63:262142":  # cast to any listener
        assert flow_marker == BIND_REQUEST_EXPECTED
        flow_marker = BIND_OFFER_EXPECTED

    elif msg._pkt._hdr == f"1FC9| W|{requestor_id}":  # sent to the requestor
        assert flow_marker == BIND_OFFER_EXPECTED
        flow_marker = BIND_CONFIRM_EXPECTED

    elif msg._pkt._hdr == f"1FC9| I|{listener_id}":  # sent to the listener/offerer
        assert flow_marker == BIND_CONFIRM_EXPECTED
        flow_marker = BIND_COMPLETED

    else:
        assert False, msg


@abort_if_rf_test_fails
async def test_hvac_bind_rem(test_port):
    """Bind a REM (remote) to a FAN (ventilation unit)."""

    global flow_marker

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, rem.id, fan.id, Code._22F1)

    config = {
        SZ_CONFIG: {SZ_DISABLE_DISCOVERY: False, SZ_ENFORCE_KNOWN_LIST: True},
        SZ_ORPHANS_HVAC: [FAN_ID, REM_ID],
        SZ_KNOWN_LIST: {
            FAN_ID: {SZ_CLASS: DevType.FAN},
            REM_ID: {SZ_CLASS: DevType.REM, SZ_FAKED: True},
        },
    }

    gwy = await load_test_gwy(*test_port, None, devices=[], **config)
    gwy.create_client(track_packet_flow_wrapper)

    fan: HvacVentilator = gwy.device_by_id[FAN_ID]

    # make an unfakeable device be fakeable...
    fan.__class__ = HvacVentilatorFakable
    setattr(fan, "_faked", None)
    setattr(fan, "_1fc9_state", {"state": _BindStates.IS_IDLE_DEVICE})

    # then enable faking on this device
    fan._make_fake()
    fan._bind_waiting(Code._22F1)

    rem: HvacRemote = gwy.device_by_id[REM_ID]

    flow_marker = BIND_REQUEST_EXPECTED
    rem._bind()

    dtm = dt.now() + td(seconds=5)
    while dtm > dt.now():
        await asyncio.sleep(0.002)
        if flow_marker == BIND_COMPLETED:
            break

    await gwy.stop()
    assert flow_marker == BIND_COMPLETED


@abort_if_rf_test_fails
async def test_hvac_bind_co2(test_port):
    """Bind a CO2 (CO2 sensor) to a FAN (ventilation unit)."""

    global flow_marker

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, co2.id, fan.id, Code._1298)

    config = {
        SZ_CONFIG: {SZ_DISABLE_DISCOVERY: False, SZ_ENFORCE_KNOWN_LIST: True},
        SZ_ORPHANS_HVAC: [FAN_ID, CO2_ID],
        SZ_KNOWN_LIST: {
            FAN_ID: {SZ_CLASS: DevType.FAN},
            CO2_ID: {SZ_CLASS: DevType.CO2, SZ_FAKED: True},
        },
    }

    gwy = await load_test_gwy(*test_port, None, devices=[], **config)
    gwy.create_client(track_packet_flow_wrapper)

    fan: HvacVentilator = gwy.device_by_id[FAN_ID]

    # make an unfakeable device be fakeable...
    fan.__class__ = HvacVentilatorFakable
    setattr(fan, "_faked", None)
    setattr(fan, "_1fc9_state", {"state": _BindStates.IS_IDLE_DEVICE})

    # then enable faking on this device
    fan._make_fake()
    fan._bind_waiting(Code._1298)

    co2: HvacRemote = gwy.device_by_id[CO2_ID]

    flow_marker = BIND_REQUEST_EXPECTED
    co2._bind()

    dtm = dt.now() + td(seconds=5)
    while dtm > dt.now():
        await asyncio.sleep(0.002)
        if flow_marker == BIND_COMPLETED:
            break

    await gwy.stop()
    assert flow_marker == BIND_COMPLETED



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test CH/DHW schedules with a mocked controller.
"""

from copy import deepcopy
from unittest.mock import patch

from tests_deprecated.common import (
    TEST_DIR,
    abort_if_rf_test_fails,
    find_test_tcs,
    load_test_gwy,
    test_ports,
)
from tests_deprecated.mocked_rf import MOCKED_PORT

from ramses_rf.const import SZ_SCHEDULE, SZ_TOTAL_FRAGS, SZ_ZONE_IDX, Code
from ramses_rf.system import DhwZone, System, Zone
from ramses_rf.system.schedule import (
    DAY_OF_WEEK,
    ENABLED,
    HEAT_SETPOINT,
    SCH_SCHEDULE_DHW,
    SCH_SCHEDULE_ZON,
    SWITCHPOINTS,
    TIME_OF_DAY,
)
from ramses_tx import Message

WORK_DIR = f"{TEST_DIR}/configs"
CONFIG_FILE = "config_heat.json"


def pytest_generate_tests(metafunc):
    metafunc.parametrize("test_port", test_ports.items(), ids=test_ports.keys())


# NOTE: used as a global
_global_flow_marker: int = None  # type: ignore[assignment]


MIN_GAP_BETWEEN_WRITES = 0  # patch ramses_tx.transport
WAITING_TIMEOUT_SECS = 0  # # patch ramses_rf.binding_fsm


RQ_0006_EXPECTED = 20
RP_0006_EXPECTED = 22
RP_0006_RECEIVED = 29

RQ_0404_FIRST_EXPECTED = 40
RP_0404_FIRST_EXPECTED = 42
RQ_0404_OTHER_EXPECTED = 44
RP_0404_OTHER_EXPECTED = 46
RP_0404_FINAL_RECEIVED = 49

W__0404_FIRST_EXPECTED = 60
I__0404_FIRST_EXPECTED = 62
W__0404_OTHER_EXPECTED = 64
I__0404_OTHER_EXPECTED = 66
I__0404_FINAL_RECEIVED = 69


def track_packet_flow(msg, tcs_id, zone_idx=None):
    """Test the flow of packets (messages)."""

    global _global_flow_marker

    if msg.code not in (Code._0006, Code._0404):
        return

    # get the schedule version number
    if msg._pkt._hdr == f"0006|RQ|{tcs_id}":
        assert _global_flow_marker == RQ_0006_EXPECTED
        _global_flow_marker = RP_0006_EXPECTED

    elif msg._pkt._hdr == f"0006|RP|{tcs_id}":
        assert _global_flow_marker == RP_0006_EXPECTED
        _global_flow_marker = RP_0006_RECEIVED  # RQ_0404_FIRST_EXPECTED

    # get the first schedule fragment, is possibly the last fragment too
    elif msg._pkt._hdr == f"0404|RQ|{tcs_id}|{zone_idx}01":
        assert _global_flow_marker in (RQ_0404_FIRST_EXPECTED, RP_0006_RECEIVED)
        _global_flow_marker = RP_0404_FIRST_EXPECTED

    elif msg._pkt._hdr == f"0404|RP|{tcs_id}|{zone_idx}01":
        assert _global_flow_marker == RP_0404_FIRST_EXPECTED
        if msg.payload["frag_number"] < msg.payload["total_frags"]:
            _global_flow_marker = RQ_0404_OTHER_EXPECTED
        else:
            _global_flow_marker = RP_0404_FINAL_RECEIVED

    # get the subsequent schedule fragments, until the last fragment
    elif msg._pkt._hdr[:20] == f"0404|RQ|{tcs_id}|{zone_idx}":
        assert _global_flow_marker == RQ_0404_OTHER_EXPECTED
        _global_flow_marker = RP_0404_OTHER_EXPECTED

    elif msg._pkt._hdr[:20] == f"0404|RP|{tcs_id}|{zone_idx}":
        assert _global_flow_marker == RP_0404_OTHER_EXPECTED
        if msg.payload["frag_number"] < msg.payload["total_frags"]:
            _global_flow_marker = RQ_0404_OTHER_EXPECTED
        else:
            _global_flow_marker = RP_0404_FINAL_RECEIVED

    # set the first schedule fragment, is possibly the last fragment too
    elif msg._pkt._hdr == f"0404| W|{tcs_id}|{zone_idx}01":
        assert _global_flow_marker in (
            W__0404_FIRST_EXPECTED,
            RP_0006_RECEIVED,
            RP_0404_FINAL_RECEIVED,
        )
        _global_flow_marker = I__0404_FIRST_EXPECTED

    elif msg._pkt._hdr == f"0404| I|{tcs_id}|{zone_idx}01":
        assert _global_flow_marker == I__0404_FIRST_EXPECTED
        if msg.payload["frag_number"] < msg.payload["total_frags"]:
            _global_flow_marker = W__0404_OTHER_EXPECTED
        else:
            _global_flow_marker = I__0404_FINAL_RECEIVED

    # set the subsequent schedule fragments, until the last fragment
    elif msg._pkt._hdr[:20] == f"0404| W|{tcs_id}|{zone_idx}":
        assert _global_flow_marker == W__0404_OTHER_EXPECTED
        _global_flow_marker = I__0404_OTHER_EXPECTED

    elif msg._pkt._hdr[:20] == f"0404| I|{tcs_id}|{zone_idx}":
        assert _global_flow_marker == I__0404_OTHER_EXPECTED
        if msg.payload["frag_number"] < msg.payload["total_frags"]:
            _global_flow_marker = W__0404_OTHER_EXPECTED
        else:
            _global_flow_marker = I__0404_FINAL_RECEIVED

    else:
        assert False, msg


def assert_schedule_dict(zone: DhwZone | Zone):
    schedule_full = zone._schedule._schedule

    assert schedule_full[SZ_ZONE_IDX] == zone.idx
    assert schedule_full[SZ_SCHEDULE] == zone.schedule

    if schedule_full[SZ_ZONE_IDX] == "HW":
        SCH_SCHEDULE_DHW(schedule_full)
    else:
        SCH_SCHEDULE_ZON(schedule_full)

    schedule = schedule_full[SZ_SCHEDULE]
    # assert isinstance(schedule, list)
    assert len(schedule) == 7

    for idx, day_of_week in enumerate(schedule):
        # assert isinstance(day_of_week, dict)
        assert day_of_week[DAY_OF_WEEK] == idx

        # assert isinstance(day_of_week[SWITCHPOINTS], dict)
        for switchpoint in day_of_week[SWITCHPOINTS]:
            assert isinstance(switchpoint[TIME_OF_DAY], str)
            if HEAT_SETPOINT in switchpoint:
                assert isinstance(switchpoint[HEAT_SETPOINT], float)
            else:
                assert isinstance(switchpoint[ENABLED], bool)

    return schedule


async def read_schedule(zone: DhwZone | Zone) -> list:  # uses: flow_marker
    """Test the get_schedule() method for a Zone or for DHW."""

    # [{  'day_of_week': 0,
    #     'switchpoints': [{'time_of_day': '06:30', 'heat_setpoint': 21.0}, ...], }]

    global _global_flow_marker

    _global_flow_marker = RQ_0006_EXPECTED
    schedule = await zone.get_schedule()  # RQ|0404, may: TimeoutError
    assert _global_flow_marker == RP_0404_FINAL_RECEIVED

    if not schedule:  # valid test?
        assert zone._msgs[Code._0404].payload[SZ_TOTAL_FRAGS] is None
        return []

    schedule = assert_schedule_dict(zone)

    zone._gwy._disable_sending = True

    _global_flow_marker = RQ_0006_EXPECTED
    assert schedule == await zone.get_schedule(force_io=False)
    assert _global_flow_marker == RQ_0006_EXPECTED

    try:
        await zone.get_schedule(force_io=True)
    except RuntimeError:  # sending is disabled
        assert True
    else:
        assert False

    return schedule


async def read_schedule_ver(tcs: System) -> list:  # uses: flow_marker
    """Test the get_schedule() method for a Zone or for DHW."""

    global _global_flow_marker

    _global_flow_marker = RQ_0006_EXPECTED
    ver = (await tcs._schedule_version())[0]  # RQ|0006, may: TimeoutError
    assert _global_flow_marker == RP_0006_RECEIVED

    assert isinstance(ver, int)
    assert ver == tcs._msgs[Code._0006].payload["change_counter"]

    _global_flow_marker = RQ_0006_EXPECTED  # actually, is not expected
    assert ver == (await tcs._schedule_version(force_io=False))[0]
    assert _global_flow_marker == RQ_0006_EXPECTED

    ver = (await tcs._schedule_version(force_io=True))[0]  # RQ|0006, may: TimeoutError
    assert _global_flow_marker == RP_0006_RECEIVED

    tcs._gwy._disable_sending = True  # TODO: must speak directly to lower layer?

    _global_flow_marker = RQ_0006_EXPECTED  # actually, is not expected
    ver = (await tcs._schedule_version())[0]  # RQ|0006, may: TimeoutError
    assert _global_flow_marker == RQ_0006_EXPECTED

    try:
        await tcs._schedule_version(force_io=True)
    except RuntimeError:  # sending is disabled
        pass
    else:
        assert False


async def write_schedule(zone: DhwZone | Zone) -> None:  # uses: flow_marker
    """Test the set_schedule() method for a Zone or for DHW."""

    # FYI: [{  'day_of_week': 0,
    #     'switchpoints': [{'time_of_day': '06:30', 'heat_setpoint': 21.0}, ...], }]

    global _global_flow_marker

    _global_flow_marker = RQ_0006_EXPECTED  # because of force_io=True
    ver_old, _ = await zone.tcs._schedule_version(force_io=True)
    assert _global_flow_marker == RP_0006_RECEIVED

    sch_old = await zone.get_schedule()
    assert _global_flow_marker == RP_0404_FINAL_RECEIVED

    sch_new = deepcopy(sch_old)

    # if zone._gwy.pkt_transport.serial.port == MOCKED_PORT:
    #     # change the schedule (doesn't matter to what)
    #     if zone.idx == "HW":
    #         sch_new[0][SWITCHPOINTS][0][ENABLED] = not (
    #             sch_new[0][SWITCHPOINTS][0][ENABLED]
    #         )
    #     else:
    #         sch_new[0][SWITCHPOINTS][0][HEAT_SETPOINT] = (
    #             sch_new[0][SWITCHPOINTS][0][HEAT_SETPOINT]
    #         ) % 30 + 5

    _ = await zone.set_schedule(sch_new)  # check zone._schedule._schedule
    assert _global_flow_marker == I__0404_FINAL_RECEIVED

    _global_flow_marker = RQ_0006_EXPECTED  # because of force_io=True
    ver_tst, _ = await zone.tcs._schedule_version(force_io=True)  # TODO: force_io=False
    assert _global_flow_marker == RP_0006_RECEIVED

    assert ver_tst > ver_old

    _global_flow_marker = RQ_0006_EXPECTED
    sch_tst = await zone.get_schedule()  # will use latest I/RP|0006
    assert _global_flow_marker == RQ_0006_EXPECTED

    sch_tst = await zone.get_schedule(force_io=True)  # will force RQ|0006
    assert _global_flow_marker == RP_0006_RECEIVED

    assert sch_tst == sch_new
    # if zone._gwy.pkt_transport.serial.port == MOCKED_PORT:
    #     assert sch_tst != sch_old
    #    sch_end = await zone.set_schedule(sch_old)  # put things back


@abort_if_rf_test_fails  # TODO: should be ramses_tx.protocol.???
@patch("ramses_rf.binding_fsm.WAITING_TIMEOUT_SECS", WAITING_TIMEOUT_SECS)
async def test_rq_0006_ver(test_port):
    """Test the TCS._schedule_version() method."""

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, tcs.id)

    gwy = await load_test_gwy(*test_port, f"{WORK_DIR}/{CONFIG_FILE}")
    gwy.create_client(track_packet_flow_wrapper)

    tcs: System = find_test_tcs(gwy)

    try:
        await read_schedule_ver(tcs)
    finally:
        await gwy.stop()


@abort_if_rf_test_fails
@patch("ramses_tx.transport.MIN_GAP_BETWEEN_WRITES", MIN_GAP_BETWEEN_WRITES)
async def test_rq_0404_dhw(test_port):
    """Test the dhw.get_schedule() method."""

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, tcs.id, dhw.idx)

    gwy = await load_test_gwy(*test_port, f"{WORK_DIR}/{CONFIG_FILE}")
    gwy.create_client(track_packet_flow_wrapper)

    tcs: System = find_test_tcs(gwy)
    dhw: DhwZone = tcs.dhw

    try:
        if dhw or test_port[0] == MOCKED_PORT:  # mocked port should have DHW
            await read_schedule(dhw)
    finally:
        await gwy.stop()


@abort_if_rf_test_fails
@patch("ramses_tx.transport.MIN_GAP_BETWEEN_WRITES", MIN_GAP_BETWEEN_WRITES)
async def test_rq_0404_zon(test_port):
    """Test the zone.get_schedule() method."""

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, tcs.id, zon.idx)

    gwy = await load_test_gwy(*test_port, f"{WORK_DIR}/{CONFIG_FILE}")
    gwy.create_client(track_packet_flow_wrapper)

    tcs: System = find_test_tcs(gwy)
    zon: Zone = tcs.zones[0]

    try:
        # if zon:
        await read_schedule(zon)
    finally:
        await gwy.stop()


@abort_if_rf_test_fails
@patch("ramses_tx.transport.MIN_GAP_BETWEEN_WRITES", MIN_GAP_BETWEEN_WRITES)
async def test_ww_0404_dhw(test_port):
    """Test the dhw.set_schedule() method (uses get_schedule)."""

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, tcs.id, dhw.idx)

    gwy = await load_test_gwy(*test_port, f"{WORK_DIR}/{CONFIG_FILE}")
    gwy.create_client(track_packet_flow_wrapper)

    tcs: System = find_test_tcs(gwy)
    dhw: DhwZone = tcs.dhw

    try:
        if dhw or test_port[0] == MOCKED_PORT:  # mocked port should have DHW
            await write_schedule(dhw)
    finally:
        await gwy.stop()


@abort_if_rf_test_fails
@patch("ramses_tx.transport.MIN_GAP_BETWEEN_WRITES", MIN_GAP_BETWEEN_WRITES)
async def test_ww_0404_zon(test_port):
    """Test the zone.set_schedule() method (uses get_schedule)."""

    def track_packet_flow_wrapper(msg: Message, *args, **kwargs):
        track_packet_flow(msg, tcs.id, zon.idx)

    gwy = await load_test_gwy(*test_port, f"{WORK_DIR}/{CONFIG_FILE}")
    gwy.create_client(track_packet_flow_wrapper)

    tcs: System = find_test_tcs(gwy)
    zon: Zone = tcs.zones[0]

    try:
        # if zon:
        await write_schedule(zon)
    finally:
        await gwy.stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the Schedule functions.
"""

import asyncio
import json
import unittest

from ramses_rf import Gateway
from ramses_rf.discovery import SET_SCHED, spawn_scripts
from tests.helpers import TEST_DIR

WORK_DIR = f"{TEST_DIR}/schedules"


class TestSchedule(unittest.IsolatedAsyncioTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.maxDiff = None
        self.gwy = None

    async def test_schedule_set(self):
        with open(f"{TEST_DIR}/schemas/schema_000.log") as f:
            self.gwy = Gateway(
                None, input_file=f, config={}, loop=self._asyncioTestLoop
            )
            await self.gwy.start()

        with open(f"{WORK_DIR}/schedule.json") as f:
            schedule = json.load(f)

        with open(f"{WORK_DIR}/schedule.json") as f:
            await asyncio.gather(
                *spawn_scripts(self.gwy, **{SET_SCHED: ("01:123456", f)})
            )

        zone = self.gwy.system_by_id["01:145038"].zone_by_idx["01"]
        result = {"zone_idx": zone.idx, "schedule": zone.schedule}

        self.assertEqual(result, schedule)

        await self.gwy.stop()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the configuration parsers.
"""


import pytest
import voluptuous as vol
import yaml

from ramses_rf.schemas import (
    SCH_GATEWAY_DICT,
    SCH_GLOBAL_SCHEMAS,
    SCH_GLOBAL_SCHEMAS_DICT,
    SCH_RESTORE_CACHE_DICT,
)
from ramses_tx.schemas import (
    SCH_ENGINE_DICT,
    SCH_GLOBAL_TRAITS_DICT,
    sch_packet_log_dict_factory,
    sch_serial_port_dict_factory,
)

SCH_PACKET_LOG_DICT = sch_packet_log_dict_factory(default_backups=7)
SCH_SERIAL_PORT_DICT = sch_serial_port_dict_factory()

SCH_GATEWAY = vol.Schema(SCH_GATEWAY_DICT | SCH_ENGINE_DICT, extra=vol.PREVENT_EXTRA)
SCH_GLOBAL_TRAITS = vol.Schema(SCH_GLOBAL_TRAITS_DICT, extra=vol.PREVENT_EXTRA)
SCH_PACKET_LOG = vol.Schema(SCH_PACKET_LOG_DICT, extra=vol.PREVENT_EXTRA)
SCH_RESTORE_CACHE = vol.Schema(SCH_RESTORE_CACHE_DICT, extra=vol.PREVENT_EXTRA)
SCH_SERIAL_PORT = vol.Schema(SCH_SERIAL_PORT_DICT, extra=vol.PREVENT_EXTRA)


# TODO: These schema pass testing, but shouldn't

_FAIL_BUT_VALID = (
    """
    01:333333: {is_vcs: true}
    """,
    """
    02:333333: {is_vcs: true}
    """,
)
_PASS_BUT_INVALID = (
    """
    main_tcs: 01:111111  # no TCS schema
    """,
    """
    known_list:
      01:111111: {}
    block_list:
      01:111111: {}  # also in known_list
    """,
)


def no_duplicates_constructor(loader, node, deep=False):
    """Check for duplicate keys."""
    mapping = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        if key in mapping:
            raise yaml.constructor.ConstructorError(
                f"Duplicate key: {key} ('{mapping[key]}' overwrites '{value_node}')"
            )
        value = loader.construct_object(value_node, deep=deep)
        mapping[key] = value
    return loader.construct_mapping(node, deep)


class CheckForDuplicatesLoader(yaml.Loader):
    """Local class to prevent pollution of global yaml.Loader."""

    pass


CheckForDuplicatesLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, no_duplicates_constructor
)


def _test_schema(validator: vol.Schema, schema: str) -> dict:
    return validator(yaml.load(schema, CheckForDuplicatesLoader))
    # return validator(yaml.safe_load(schema))  # PyYAML silently swallows duplicate keys!


def _test_schema_bad(validator: vol.Schema, schema: str) -> None:
    global test_schemas_bad_failed
    try:
        _test_schema(validator, schema)
    except (vol.MultipleInvalid, yaml.YAMLError):
        pass
    else:
        test_schemas_bad_failed = True
        raise TypeError(f"should *not* be valid YAML, but parsed OK: {schema}")


def _test_schema_good(validator: vol.Schema, schema: str) -> dict:
    global test_schemas_good_failed
    try:
        _test_schema(validator, schema)
    except (vol.MultipleInvalid, yaml.YAMLError) as exc:
        test_schemas_good_failed = True
        raise TypeError(f"should be valid YAML, but didn't parse OK: {schema}") from exc


GATEWAY_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    disable_discovery: null
    """,
    """
    max_zones: 19
    """,
    """
    use_regex:
    """,
    """
    disable_discovery: null
    enable_eavesdrop: null
    max_zones: null
    reduce_processing: null
    use_aliases: null
    use_native_ot: null
    """,
)
GATEWAY_GOOD = (
    """
    {}
    """,
    """
    disable_discovery: false
    """,
    """
    max_zones: 16
    """,
    """
    disable_discovery: false
    enable_eavesdrop: false
    max_zones: 12
    reduce_processing: 0
    use_aliases: false
    use_native_ot: prefer
    """,
    """
    disable_discovery: true
    enable_eavesdrop: true
    max_zones: 3
    reduce_processing: 2
    use_aliases: true
    use_native_ot: never
    use_regex: {}
    """,
    """
    disable_sending: true
    enforce_known_list: true
    evofw_flag: "!V"
    use_regex:
      inbound: {}
      outbound: {}
    """,
)


@pytest.mark.parametrize("index", range(len(GATEWAY_BAD)))
def test_gateway_bad(index, schemas=GATEWAY_BAD):
    _test_schema_bad(SCH_GATEWAY, schemas[index])


@pytest.mark.parametrize("index", range(len(GATEWAY_GOOD)))
def test_gateway_good(index, schemas=GATEWAY_GOOD):
    _test_schema_good(SCH_GATEWAY, schemas[index])


KNOWN_LIST_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    known_list: []  # expected a dictionary for dictionary value @ data['known_list']
    """,
    """
    known_list:
      01:111111: {class: xxx}
    """,
    # """
    # known_list:
    #   01:111111: {class: CTL, notes: this is invalid _note}
    #   02:111111: {class: UFC, extra: this is invalid too}
    # """,
    """
    known_list:
      01:111111: {class: CTL}
      01:111111: {class: UFC}  # Duplicate key: 01:111111
    """,
    # """
    # known_list:
    #   05:111111: {class: REM, scheme: xxxxxx}
    # """,
)
KNOWN_LIST_GOOD = (
    """
    {}
    """,
    """
    known_list: {}
    """,
    """
    known_list: {}
    block_list: {}
    """,
    """
    known_list:
      01:111111: {}
    block_list:
      01:222222: {}
    """,
    """
    known_list:
      01:111111: {class: CTL}
      02:111111: {class: UFC}
      03:111111: {class: THM, faked: true}
      04:111111: {class: TRV}
      07:111111: {class: DHW, faked: true}
      10:111111: {class: OTB}
      12:111111: {class: THM}
      13:111111: {class: BDR}
      17:111111: {class: OUT, faked: true}
      18:111111: {class: HGI}
      22:111111: {class: THM}
      23:111111: {class: THM}
      30:111111: {class: RFG}
      34:111111: {class: THM, _note: this is a note}
    """,
    """
    known_list:
      01:111111: {class: FAN}
      02:111111: {class: RFS}
      03:111111: {class: CO2, faked: true}
      04:111111: {class: HUM, faked: true}
      05:111111: {class: REM, faked: true, scheme: nuaire}
      06:111111:
        class: DIS
        scheme: orcon
        _note: this is a note
    """,
)


@pytest.mark.parametrize("index", range(len(KNOWN_LIST_BAD)))
def test_known_list_bad(index, schemas=KNOWN_LIST_BAD):
    _test_schema_bad(SCH_GLOBAL_TRAITS, schemas[index])


@pytest.mark.parametrize("index", range(len(KNOWN_LIST_GOOD)))
def test_known_list_good(index, schemas=KNOWN_LIST_GOOD):
    _test_schema_good(SCH_GLOBAL_TRAITS, schemas[index])


PACKET_LOG_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    packet_log:
      file_name: null  # expected str for dictionary value @ data['packet_log']['file_name']
    """,
    """
    packet_log:  # required key not provided @ data['packet_log']['file_name']
      rotate_backups: 7
      rotate_bytes: 204800
    """,
)
PACKET_LOG_GOOD = (
    """
    {}
    """,
    """
    packet_log: packet.log
    """,
    """
    packet_log: null  # expected str for dictionary value @ data['packet_log']
    """,
    """
    packet_log:
      file_name: packet.log
    """,
    """
    packet_log:
      file_name: packet.log
      rotate_backups: 7
    """,
    """
    packet_log:
      file_name: packet.log
      rotate_bytes: 204800
    """,
    """
    packet_log:
      file_name: packet.log
      rotate_backups: 7
      rotate_bytes: 204800
    """,
)


@pytest.mark.parametrize("index", range(len(PACKET_LOG_BAD)))
def test_packet_log_bad(index, schemas=PACKET_LOG_BAD):
    _test_schema_bad(SCH_PACKET_LOG, schemas[index])


@pytest.mark.parametrize("index", range(len(PACKET_LOG_GOOD)))
def test_packet_log_good(index, schemas=PACKET_LOG_GOOD):
    _test_schema_good(SCH_PACKET_LOG, schemas[index])


RESTORE_CACHE_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    restore_cache: none  # should be boolean
    """,
    """
    restore_cache: true
      restore_schema: true  # yaml.scanner.ScannerError
    """,
    """
    restore_schema: true  # should be: restore_cache: restore_schema: true
    """,
    """
    restore_state: false  # should be: restore_cache: restore_state: true
    """,
)
RESTORE_CACHE_GOOD = (
    """
    {}
    """,
    """
    restore_cache: false
    """,
    """
    restore_cache: true
    """,
    """
    restore_cache:
      restore_schema: true
    """,
    """
    restore_cache:
      restore_state:  true
    """,
    """
    restore_cache:
      restore_schema: true
      restore_state:  false
    """,
    """
    restore_cache:
      restore_schema: false
      restore_state:  true
    """,
)


@pytest.mark.parametrize("index", range(len(RESTORE_CACHE_BAD)))
def test_restore_cache_bad(index, schemas=RESTORE_CACHE_BAD):
    _test_schema_bad(SCH_RESTORE_CACHE, schemas[index])


@pytest.mark.parametrize("index", range(len(RESTORE_CACHE_GOOD)))
def test_restore_cache_good(index, schemas=RESTORE_CACHE_GOOD):
    _test_schema_good(SCH_RESTORE_CACHE, schemas[index])


SERIAL_PORT_BAD = (
    """
    #  expected a dictionary
    """,
    """
    {}  # required key not provided @ data['serial_port']
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    serial_name: /dev/ttyMOCK  # should be: serial_port:
    """,
    """
    serial_port: /dev/ttyMOCK  # yaml.parser.ParserError
      baudrate: 115200  # default
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
      baud_rate: 57600  # should be: baudrate:
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
        baudrate: 57600  # yaml.parser.ScannerError
    """,
)
SERIAL_PORT_GOOD = (
    """
    serial_port: /dev/ttyMOCK
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
      baudrate: 115200  # default
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
      baudrate: 57600
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
      baudrate: 57600
    """,
    """
    serial_port:
      port_name: /dev/ttyMOCK
      baudrate: 57600
      dsrdtr: false
      rtscts: false
      timeout: 0
      xonxoff: true
    """,
)


@pytest.mark.parametrize("index", range(len(SERIAL_PORT_BAD)))
def test_serial_port_bad(index, schemas=SERIAL_PORT_BAD):
    _test_schema_bad(SCH_SERIAL_PORT, schemas[index])


@pytest.mark.parametrize("index", range(len(SERIAL_PORT_GOOD)))
def test_serial_port_good(index, schemas=SERIAL_PORT_GOOD):
    _test_schema_good(SCH_SERIAL_PORT, schemas[index])


SCHEMAS_TCS_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    01:111111:  # expected a dictionary for dictionary value @ data['01:111111']
    """,
    """
    01:111111:
      system:  # expected a dictionary for dictionary value @ data['01:111111']['system']
    """,
    """
    13:111111:  # should be: 01:111111
      system: {}  # The ventilation control system schema must include at least one of [remotes, sensors] @ data['13:111111']
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
      zones:  # should be "00"
        00: {}  # extra keys not allowed @ data['01:111111']['zones'][0]
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
      zones:
        "00":  # extra keys not allowed @ data['01:111111']['zones']['00']
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
      zones:
        "1C": {}  # extra keys not allowed @ data['01:111111']['zones']['1C']
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
    01:111111: {remotes: [29:111111, 29:222222]}
    """,
)
SCHEMAS_TCS_GOOD = (
    """
    {}
    """,
    """
    01:111111: {}
    """,
    """
    01:111111: {is_tcs: true}
    """,
    """
    01:111111:
      system: {}
    """,
    """
    01:111111:
      system:
        appliance_control:
    """,
    """
    01:111111:
      system:
        appliance_control: null
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
    """,
    """
    01:111111:
      system:
        appliance_control: 10:111111
    01:222222:
      system:
        appliance_control: 13:222222
    """,
    """
    main_tcs: null
    01:111111:
      system:
        appliance_control: 10:111111
      zones:
        "0B":
          sensor: 01:111111
    """,
    """
    main_tcs: 01:222222
    01:111111:
      system:
        appliance_control: 10:111111
      zones:
        "00": {}
        "01": {sensor: 03:111111}
        "02": {actuators: [04:111111, 04:222222]}
        "03": {actuators: [13:111111, 13:222222]}
    """,
    """
    main_tcs: 01:222222
    01:111111:
      system:
        appliance_control: 10:111111
    01:222222:
      system:
        appliance_control: 10:222222
    """,
)


@pytest.mark.parametrize("index", range(len(SCHEMAS_TCS_BAD)))
def test_schemas_tcs_bad(index, schemas=SCHEMAS_TCS_BAD):
    _test_schema_bad(SCH_GLOBAL_SCHEMAS, schemas[index])


@pytest.mark.parametrize("index", range(len(SCHEMAS_TCS_GOOD)))
def test_schemas_tcs_good(index, schemas=SCHEMAS_TCS_GOOD):
    _test_schema_good(SCH_GLOBAL_SCHEMAS, schemas[index])


SCHEMAS_VCS_BAD = (
    """
    #  expected a dictionary
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
    32:111111:  # expected a dictionary for dictionary value @ data['01:111111']
    """,
    """
    32:111111: {is_vcs: true}
    """,
    """
    01:111111:
      remotes: []
      is_tcs: true
    """,
    """
    32:111111:  # should not duplicate device_id
      remotes: [29:111111, 29:111111]  # not a valid value for dictionary value @ data['32:111111']['remotes']
    """,
    """
    32:111111: {remotes: [29:111111, 29:222222]}
    32:111111: {remotes: [29:111111, 29:222222]}  # has duplicate key
    32:333333: {remotes: [29:111111, 29:222222]}
    """,
)
SCHEMAS_VCS_GOOD = (
    """
    {}
    """,
    """
    01:333333:
      remotes: []
    """,
    """
    32:111111:
      remotes: []
    """,
    """
    32:111111:
      remotes: []
      is_vcs: true
    """,
    """
    32:111111:
      remotes: [29:111111]
    """,
    """
    32:111111:
      remotes: [29:111111, 29:222222]
      sensors: [29:111111, 29:333333]
      is_vcs: true
    """,
    """
    32:111111: {remotes: [29:111111, 29:222222]}
    32:222222: {remotes: [29:111111, 29:222222]}
    32:333333: {remotes: [29:111111, 29:222222]}
    """,
)


@pytest.mark.parametrize("index", range(len(SCHEMAS_VCS_BAD)))
def test_schemas_vcs_bad(index, schemas=SCHEMAS_VCS_BAD):
    _test_schema_bad(SCH_GLOBAL_SCHEMAS, schemas[index])


@pytest.mark.parametrize("index", range(len(SCHEMAS_VCS_GOOD)))
def test_schemas_vcs_good(index, schemas=SCHEMAS_VCS_GOOD):
    _test_schema_good(SCH_GLOBAL_SCHEMAS, schemas[index])


SCHEMAS_MIXED_BAD = tuple(x + y for x in SCHEMAS_TCS_GOOD for y in SCHEMAS_VCS_BAD[1:])
SCHEMAS_MIXED_BAD += tuple(
    x + y for x in SCHEMAS_TCS_BAD[1:] for y in SCHEMAS_VCS_GOOD[1:]
)
SCHEMAS_MIXED_GOOD = tuple(
    x + y for x in SCHEMAS_TCS_GOOD[1:] for y in SCHEMAS_VCS_GOOD[1:]
)


test_schemas_bad_failed = False
test_schemas_good_failed = False


@pytest.mark.parametrize("index", range(len(SCHEMAS_MIXED_BAD)))
def test_schemas_mixed_bad(index, schemas=SCHEMAS_MIXED_BAD):
    global test_schemas_bad_failed
    if not test_schemas_bad_failed:
        _test_schema_bad(SCH_GLOBAL_SCHEMAS, schemas[index])


@pytest.mark.parametrize("index", range(len(SCHEMAS_MIXED_GOOD)))
def test_schemas_mixed_good(index, schemas=SCHEMAS_MIXED_GOOD):
    global test_schemas_good_failed
    if not test_schemas_good_failed:
        _test_schema_good(SCH_GLOBAL_SCHEMAS, schemas[index])


SCHEMAS_HASS_BAD = (
    """
    #  expected a dictionary
    """,
    """
    {}
    """,
    """
    other_key: null  # extra keys not allowed @ data['other_key']
    """,
    """
ramses_rf:
  serial_port: /dev/ttyUSB0

  ramses_rf:
    enforce_known_list: true
    """,
    """
ramses_cc:
  serial_port: /dev/ttyUSB0

  ramses_cc:
    enforce_known_list: true
    """,
    """
ramses_cc:
  serial_port: /dev/ttyUSB0

  schema:  # this not used
    01:111111:  # Temperature control system (e.g. evohome)
      system:
        appliance_control: 10:123446
      zones:
        "07": {sensor: 01:111111}

    orphans_hvac: [30:111111, 32:333333, 32:555555, 32:666666]
    """,
    """
ramses_cc:
  serial_port: /dev/ttyACM0
  01:123456: {}
  01:123456: {}  # Duplicate key: 01:123456...
    """,
)
SCHEMAS_HASS_GOOD = (
    """
ramses_cc:
  serial_port: /dev/ttyACM0
    """,
    """
ramses_cc:
  serial_port: /dev/ttyUSB0
  restore_cache: false
  known_list:
    """,
    """
ramses_cc:
  serial_port: rfc2217://localhost:5001
  restore_cache:
    restore_schema: true
  known_list: {}
  block_list:
    """,
    """
ramses_cc:
  serial_port: /dev/serial/by-id/usb-SHK_NANO_CUL_868-if00-port0
  restore_cache:
    restore_state: true
  known_list:
    30:111111:              # becomes empty {}
    32:333333: {}           #
    32:555555: {class: CO2} #
  block_list: {}
    """,
    """
ramses_cc:
  serial_port: /dev/serial/by-id/usb-SHK_NANO_CUL_868-if00-port0

  scan_interval: 300

  restore_cache:
    restore_schema: true
    restore_state: true

  packet_log:
    file_name: packet.log
    rotate_backups: 28

  ramses_rf:
    enforce_known_list: true

  main_tcs: 01:111111

  01:111111:  # Temperature control system (e.g. evohome)
    system:
      appliance_control: 10:123446
    zones:
      "07": {sensor: 01:111111}

  orphans_heat: [02:123456]
  orphans_hvac: [30:111111, 32:333333, 32:555555, 32:666666]

  known_list:
    30:111111: {class: FAN}               # Orcon HRU
    32:333333: {class: REM, faked: true}  # an impersonatable remote
    32:555555: {class: CO2, faked: true}  # a fully faked sensor
    32:666666: {class: HUM}

  block_list:
    23:111111: {}

  advanced_features:
    send_packet: true
    """,
    """
ramses_cc:
  serial_port: /dev/ttyACM0
  01:123456: {}
  01:123457: {}
    """,
)


SCH_DOMAIN_CONFIG = vol.All(  # as per ramses_cc.schemas, TODO: add advanced_features
    vol.Schema(
        {
            vol.Optional("ramses_rf", default={}): SCH_GATEWAY,
            vol.Optional("scan_interval"): int,
            vol.Optional("advanced_features", default={}): dict,
        },
        extra=vol.PREVENT_EXTRA,
    )
    .extend(SCH_GLOBAL_SCHEMAS_DICT)
    .extend(SCH_GLOBAL_TRAITS_DICT)
    .extend(SCH_PACKET_LOG_DICT)
    .extend(SCH_RESTORE_CACHE_DICT)
    .extend(SCH_SERIAL_PORT_DICT),
)

SCH_GLOBAL_HASS = vol.Schema(
    {vol.Required("ramses_cc"): SCH_DOMAIN_CONFIG}, extra=vol.PREVENT_EXTRA
)


@pytest.mark.parametrize("index", range(len(SCHEMAS_HASS_BAD)))
def test_schemas_hass_bad(index, schemas=SCHEMAS_HASS_BAD):
    _test_schema_bad(SCH_GLOBAL_HASS, schemas[index])


@pytest.mark.parametrize("index", range(len(SCHEMAS_HASS_GOOD)))
def test_schemas_hass_good(index, schemas=SCHEMAS_HASS_GOOD):
    _test_schema_good(SCH_GLOBAL_HASS, schemas[index])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the Command.put_*, Command.set_* APIs.
"""

from datetime import datetime as dt

from ramses_tx.command import CODE_API_MAP, Command
from ramses_tx.message import Message
from ramses_tx.packet import Packet


def _test_api(api, packets):  # noqa: F811  # NOTE: incl. addr_set check
    """Test a verb|code pair that has a Command constructor."""

    for pkt_line, kwargs in packets.items():
        pkt = _create_pkt_from_frame(pkt_line)

        msg = Message(pkt)

        _test_api_from_kwargs(api, pkt, **kwargs)
        _test_api_from_msg(api, msg)


def _create_pkt_from_frame(pkt_line) -> Packet:  # noqa: F811
    """Create a pkt from a pkt_line and assert their frames match."""

    pkt = Packet.from_port(dt.now(), pkt_line)
    assert str(pkt) == pkt_line[4:]
    return pkt


def _test_api_from_msg(api, msg) -> Command:  # noqa: F811
    """Create a cmd from a msg and assert they're equal (*also* asserts payload)."""

    cmd = api(
        msg.dst.id,
        src_id=msg.src.id,
        **{k: v for k, v in msg.payload.items() if k[:1] != "_"},
    )

    assert cmd == msg._pkt  # must have exact same addr set

    return cmd


def _test_api_from_kwargs(api, pkt, **kwargs):
    cmd = api(HRU, src_id=REM, **kwargs)

    assert str(cmd) == str(pkt)


def test_set():  # noqa: F811
    for test_pkts in (SET_22F1_KWARGS, SET_22F7_KWARGS):
        pkt = list(test_pkts)[0]
        api = CODE_API_MAP[f"{pkt[4:6]}|{pkt[41:45]}"]
        _test_api(api, test_pkts)


HRU = "32:155617"
REM = "37:171871"
NUL = "--:------"

SET_22F1_KWARGS = {
    f"000  I --- {REM} {HRU} {NUL} 22F1 002 0000": {"fan_mode": None},
    #
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0000": {"fan_mode": 0},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0001": {"fan_mode": 1},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0002": {"fan_mode": 2},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0003": {"fan_mode": 3},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0004": {"fan_mode": 4},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0005": {"fan_mode": 5},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0006": {"fan_mode": 6},
    f"001  I --- {REM} {HRU} {NUL} 22F1 002 0007": {"fan_mode": 7},
    #
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0000": {"fan_mode": "00"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0001": {"fan_mode": "01"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0002": {"fan_mode": "02"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0003": {"fan_mode": "03"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0004": {"fan_mode": "04"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0005": {"fan_mode": "05"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0006": {"fan_mode": "06"},
    f"002  I --- {REM} {HRU} {NUL} 22F1 002 0007": {"fan_mode": "07"},
    #
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0000": {"fan_mode": "away"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0001": {"fan_mode": "low"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0002": {"fan_mode": "medium"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0003": {"fan_mode": "high"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0004": {"fan_mode": "auto"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0005": {"fan_mode": "auto_alt"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0006": {"fan_mode": "boost"},
    f"003  I --- {REM} {HRU} {NUL} 22F1 002 0007": {"fan_mode": "off"},
}


SET_22F7_KWARGS = {
    f"000  W --- {REM} {HRU} {NUL} 22F7 002 00FF": {},  # shouldn't be OK
    #
    f"001  W --- {REM} {HRU} {NUL} 22F7 002 00FF": {
        "bypass_position": None
    },  # is auto?
    f"001  W --- {REM} {HRU} {NUL} 22F7 002 0000": {"bypass_position": 0.0},
    # 001  W --- {REM} {HRU} {NUL} 22F7 002 0064": {"bypass_position": 0.5},
    f"001  W --- {REM} {HRU} {NUL} 22F7 002 00C8": {"bypass_position": 1.0},
    f"002  W --- {REM} {HRU} {NUL} 22F7 002 00FF": {
        "bypass_mode": "auto"
    },  # is auto, or None?
    f"002  W --- {REM} {HRU} {NUL} 22F7 002 0000": {"bypass_mode": "off"},
    f"002  W --- {REM} {HRU} {NUL} 22F7 002 00C8": {"bypass_mode": "on"},
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Mocked devices used for testing.Will provide an appropriate Tx for a given Rx.
"""

from __future__ import annotations

import asyncio
import logging
from datetime import datetime as dt
from datetime import timedelta as td
from queue import Full
from typing import TYPE_CHECKING, Callable

from ramses_rf.const import I_, RP, RQ, SZ_ACTUATORS, SZ_ZONES, W_, ZON_ROLE_MAP, Code
from ramses_rf.schemas import SZ_CLASS
from ramses_tx import PacketInvalid
from ramses_tx.command import Command as CommandBase
from ramses_tx.command import validate_api_params

from .const import GWY_ID, __dev_mode__

if TYPE_CHECKING:  # mypy TypeVars and similar (e.g. Index, Verb)
    # skipcq: PY-W2000
    from ramses_tx.address import DeviceId
    from ramses_tx.frame import _PktIdxT

    from .const import Index, Verb  # noqa: F401, pylint: disable=unused-import


DEV_MODE = __dev_mode__

CTL_ID = "01:888888"
THM_ID = "03:123456"


_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


class Command(CommandBase):
    @classmethod  # constructor for RP/0005
    @validate_api_params()
    def put_zone_types(
        cls,
        src_id: str,
        dst_id: str,
        zone_type: str,
        zone_mask: tuple,
        *,
        sub_idx: str = "00",
    ):
        """Constructor for RP/0005."""

        zones = f"{sum(b<<i for i, b in enumerate(zone_mask)):04X}"
        payload = f"{sub_idx}{zone_type}{zones[2:]}{zones[:2]}"  # swap order

        return cls._from_attrs(RP, Code._0005, payload, addr0=src_id, addr1=dst_id)

    @classmethod  # constructor for RP/000C
    @validate_api_params(has_zone=True)
    def put_zone_devices(
        cls,
        src_id: str,
        dst_id: str,
        zone_idx: int | str,
        zone_type: str,
        devices: tuple[str],
    ):
        """Constructor for RP/000C."""

        payload = f"{zone_idx}{zone_type}..."

        return cls._from_attrs(RP, Code._000C, payload, addr0=src_id, addr1=dst_id)


class MockDeviceBase:
    """A pseudo-mocked device used for testing."""

    def __init__(self, gwy, device_id) -> None:
        self._gwy = gwy
        self._loop = gwy._loop
        self._ether = gwy.pkt_transport.serial._que

        self.id: DeviceId = device_id

        self._is_running: bool = None  # type: ignore[assignment]
        self._tasks: list[asyncio.Task] = []

    def rx_frame_as_cmd(self, cmd: Command) -> None:
        """Find/Create an encoded frame (via its hdr), and place in on the ether.

        The rp_header is the sent cmd's rx_header (the header of the packet the gwy is
        waiting for).
        """

        pkt_header = cmd.tx_header

        if response := RESPONSES.get(pkt_header):
            self.tx_frames_as_cmds((self.make_response_pkt(response),))

    def tx_frames_as_cmds(self, cmds: Command | tuple[Command]) -> None:
        """Queue pkts on the ether for the gwy to Rx."""

        cmds = tuple() if cmds is None else cmds  # safety net

        for cmd in cmds if isinstance(cmds, tuple) else (cmds,):
            try:
                self._ether.put_nowait((0, dt.now(), cmd))
            except Full:
                pass

    def make_response_pkt(self, frame: str) -> Command:
        """Convert the response into a command."""

        try:
            return Command(frame)
        except PacketInvalid as exc:
            raise PacketInvalid(f"Invalid entry the response table: {exc}")


_1F09_CYCLE_DURATION = td(seconds=3 * 60)  # varies between controllers, usu. 3-5 mins
_1F09_CYCLE_PERIOD = 20  # ~number of cycles until 000A is sent instead of 2309/30C9
_1F09_CYCLE_REMAINING = td(seconds=5)

_3150_CYCLE_DURATION = td(seconds=20 * 60)  # fixed at 20 mins
_3150_CYCLE_REMAINING = td(seconds=19 * 60)

DEFAULT_3B00_CYCLE_RATE = 3  # default 3/hr?
DEFAULT_3B00_CYCLE_DURATION = td(seconds=60 * 60 / DEFAULT_3B00_CYCLE_RATE)
_3B00_CYCLE_REMAINING = td(seconds=9 * 60)

DeviceIdT = str
ZoneIdT = str
ZoneIdxT = str
ZoneNameT = str
ZoneParamsT = dict[str, None | bool | float | int]
ZoneSetpointT = float
ZoneTypeT = str


SZ_HEATING_TYPE = "heating_type"
SZ_MAX_SETPOINT = "max_setpoint"
SZ_MIN_SETPOINT = "min_setpoint"
SZ_NAME = "name"
SZ_USE_EVOTOUCH = "use_evotouch"

DEFAULT_MAX_SETPOINT = 35.0
DEFAULT_MIN_SETPOINT = 5.0

SZ_ACTUATOR_RUN_TIME = "actuator_run_time"
SZ_LOCAL_OVERRIDE = "local_override"
SZ_MAX_FLOW_TEMP = "max_flow_temp"
SZ_MIN_FLOW_TEMP = "min_flow_temp"
SZ_MULTI_ROOM_MODE = "multi_room_mode"
SZ_PUMP_RUN_TIME = "pump_run_time"
SZ_WINDOW_OPEN = "window_open"

MIX_ZONE_PARAMS = (
    SZ_ACTUATOR_RUN_TIME,
    SZ_MAX_FLOW_TEMP,
    SZ_MIN_FLOW_TEMP,
    SZ_PUMP_RUN_TIME,
)
RAD_ZONE_PARAMS = (SZ_LOCAL_OVERRIDE, SZ_MULTI_ROOM_MODE, SZ_WINDOW_OPEN)

ACTUATOR_RUN_TIME_DEFAULT: int = 150  # (10-240/ secs)
LOCAL_OVERRIDE_DEFAULT: bool = False  # TODO: check
MAX_FLOW_TEMP_DEFAULT: float = 55.0  # (0.0-99.0/0.5 'C)
MIN_FLOW_TEMP_DEFAULT: float = 15.0  # (0.0-50.0/0.5 'C)
MULTI_ROOM_MODE_DEFAULT: bool = False  # TODO: check
PUMP_RUN_TIME_DEFAULT: int = 15  # (0-99/1 mins)
WINDOW_OPEN_DEFAULT = False  # TODO: check


class MockWRB:  # BDR91A/T
    pass


class MockOTB:  # R8810/20
    pass


class MockDhwZone:
    pass


class MockZone:
    """A mocked zone."""

    _name: ZoneNameT
    _heating_type: ZoneTypeT

    _max_setpoint: ZoneSetpointT
    _min_setpoint: ZoneSetpointT
    _parameters: ZoneParamsT

    _actuators: list[DeviceIdT]
    _sensor_id: None | DeviceIdT

    def __init__(
        self,
        ctl,
        zone_idx,
        name: ZoneNameT,
        heating_type: ZoneTypeT,
        use_evotouch: bool,
        **kwargs,
    ) -> None:
        """Instantiate a zone."""

        if zone_idx in ctl._zones:
            raise KeyError(f"Duplicate zone_idx: {zone_idx}")

        self._ctl = ctl
        self._idx = zone_idx

        self._actuators = []
        self._sensor_id = None

        self.name = name
        self.heating_type = heating_type
        if use_evotouch:
            self._use_evotouch(use_evotouch)  # sets self._sensor_id

        self._max_setpoint = 35.0  # initial default
        self._min_setpoint = 5.0  # initial default
        self.parameters = kwargs

    @property
    def name(self) -> ZoneNameT:
        """Return the name of the zone."""
        return self._name

    @name.setter
    def name(self, value: ZoneNameT) -> None:
        """Set the name of the zone."""

        if not isinstance(value, ZoneNameT):
            raise TypeError(f"Invalid {SZ_NAME} parameter for zone: {value}")
        self._name = value[:12]  # evotouch limits to 12, RAMSES allows more

    @property
    def heating_type(self) -> ZoneTypeT:
        """Return the heating_type of the zone."""
        return self._heating_type

    @heating_type.setter
    def heating_type(self, value: ZoneTypeT) -> None:
        """Set the heating_type of the zone (ELE, MIX, RAD, UFH, or ZON)."""

        if value == self._heating_type:
            return
        if value not in ("ELE", "MIX", "RAD", "UFH", "ZON"):
            raise TypeError(f"Invalid {SZ_HEATING_TYPE} parameter for zone: {value}")

        self._heating_type = value
        self.parameters = {}  # use setter: will set params to defaults

    @property
    def use_evotouch(self) -> bool:
        """Return True if the controller is the zone sensor."""
        return self.sensor == self._ctl.id

    def _use_evotouch(self, value: bool) -> None:
        """Set (or unset) the controller as the zone sensor.

        If the controller (even indirectly) invokes this method with value == False,
        (e.g. when instaintiating the zone) then it should immediately invoke the
        _set_sensor() method.
        """

        if value == self.use_evotouch:
            return
        if not isinstance(value, bool):
            raise TypeError(f"Invalid {SZ_USE_EVOTOUCH} parameter for zone: {value}")
        self._sensor_id = self._ctl.id if value else None

    @property
    def max_setpoint(self) -> float:
        """Return the max_setpoint of the zone (in °C)."""
        return self._max_setpoint

    @max_setpoint.setter
    def max_setpoint(self, value: None | float) -> None:
        """Set the max_setpoint of the zone (in °C).

        None is a sentinel for the default value, 35.0.
        """

        value = DEFAULT_MAX_SETPOINT if value is None else value
        if not isinstance(value, float) or not 5.0 <= value <= 35.0:
            raise TypeError(f"Invalid {SZ_MAX_SETPOINT} parameter for zone: {value}")
        self._max_setpoint = value

    @property
    def min_setpoint(self) -> float:
        """Return the min_setpoint of the zone (in °C)."""
        return self._min_setpoint

    @min_setpoint.setter
    def min_setpoint(self, value: None | float) -> None:
        """Set the min_setpoint of the zone (in °C).

        None is a sentinel for the default value, 5.0.
        """

        value = DEFAULT_MIN_SETPOINT if value is None else value
        if not isinstance(value, float) or not 5.0 <= value <= 35.0:
            raise TypeError(f"Invalid {SZ_MIN_SETPOINT} parameter for zone: {value}")
        self._min_setpoint = value

    @property
    def parameters(self) -> ZoneParamsT:
        """Return the parameters of the zone (can include min/max setpoint)."""
        return self._parameters

    @parameters.setter
    def parameters(self, params: ZoneParamsT) -> None:
        """Set the parameters of the zone (can include min/max setpoint).

        In each case, None is a sentinel for the default value.
        """

        PARAM_TABLE = {
            SZ_ACTUATOR_RUN_TIME: (
                ACTUATOR_RUN_TIME_DEFAULT,
                lambda x: isinstance(x, int) and 10 <= x <= 240,
            ),
            SZ_LOCAL_OVERRIDE: (
                LOCAL_OVERRIDE_DEFAULT,
                lambda x: isinstance(x, bool),
            ),
            SZ_MAX_FLOW_TEMP: (
                MAX_FLOW_TEMP_DEFAULT,
                lambda x: isinstance(x, float) and 0.0 <= x <= 99.0,
            ),
            SZ_MIN_FLOW_TEMP: (
                MIN_FLOW_TEMP_DEFAULT,
                lambda x: isinstance(x, float) and 0.0 <= x <= 50.0,
            ),
            SZ_MULTI_ROOM_MODE: (
                MULTI_ROOM_MODE_DEFAULT,
                lambda x: isinstance(x, bool),
            ),
            SZ_PUMP_RUN_TIME: (
                PUMP_RUN_TIME_DEFAULT,
                lambda x: isinstance(x, int) and 0 <= x <= 99,
            ),
            SZ_WINDOW_OPEN: (
                WINDOW_OPEN_DEFAULT,
                lambda x: isinstance(x, bool),
            ),
        }

        def param_value(
            param_name: str, new_params: ZoneParamsT, old_params: ZoneParamsT
        ) -> bool | float | int:
            """Return a valid parameter value."""
            default_value, is_valid = PARAM_TABLE.get[param_name]
            if param_name not in new_params:
                return old_params.get(param_name, default_value)
            elif (new_value := new_params[param_name]) is None:  # sentinel for default
                return default_value
            elif is_valid(new_value):
                return new_value  # may raise TypeError
            else:
                raise TypeError(f"Invalid parameter for zone: {param_name}={new_value}")

        def update_params_mix(old: ZoneParamsT, new: ZoneParamsT) -> ZoneParamsT:
            """Configure the parameters specific to a Mixing valve zone."""
            return {p: param_value(p, old, new) for p in MIX_ZONE_PARAMS}

        def update_params_rad(old: ZoneParamsT, new: ZoneParamsT) -> ZoneParamsT:
            """Configure the parameters specific to a Radiator zone."""
            return {p: param_value(p, old, new) for p in RAD_ZONE_PARAMS}

        if SZ_MAX_SETPOINT in params:
            self.max_setpoint = params.pop(params, SZ_MAX_SETPOINT)
        if SZ_MIN_SETPOINT in params:
            self.min_setpoint = params.pop(params, SZ_MIN_SETPOINT)

        if self.heating_type == "MIX":
            self._parameters = update_params_mix(self._parameters, params)
        elif self.heating_type == "RAD":
            self._parameters = update_params_rad(self._parameters, params)
        else:
            self._parameters = {}

    @property
    def sensor(self) -> DeviceIdT:
        """Return the zone sensor's device_id."""
        return self._sensor_id

    def _set_sensor(self, device_id: None | DeviceIdT) -> None:
        """Set a new zone sensor.

        If `device_id` is None, simply clear the current sensor.
        Otherwise, the controller should have bound with the sensor immediately before
        invoking this method (unless the controller is the sensor).
        """

        if device_id == self._sensor_id:
            return
        if not isinstance(device_id, (type(None), DeviceIdT)):
            raise TypeError(f"Invalid device_id for zone sensor: {device_id}")
        self._sensor_id = device_id

    @property
    def actuators(self) -> list[DeviceIdT]:
        """Return the zone actuators' device_ids."""
        return self._actuators

    def _add_actuator(self, device_id: DeviceIdT) -> None:
        """Add an actuator to the zone.

        The controller should have bound with the actuator immediately before invoking
        this method.
        """
        # The command code will depend upon the zone.heating_type

        if not isinstance(device_id, DeviceIdT):
            raise TypeError(f"Invalid device_id for zone actuator: {device_id}")
        self._actuators.add(device_id)


class MockDeviceCtl(MockDeviceBase):
    """A pseudo-mocked controller used for testing.

    Will periodically Rx a sync_cycle set that will be available via `read()`.
    Will use a reponse table to provide a known Rx for a given Tx sent via `write()`.
    """

    _1F09_DURATION = _1F09_CYCLE_DURATION  # varies between controllers

    _appliance_control: MockWRB | MockOTB  # TBA: | MockHPR (heat pump relay)
    _dhw_zone: MockDhwZone
    _zones: MockZone

    def __init__(
        self, gwy, device_id, *, cycle_rate: int = DEFAULT_3B00_CYCLE_RATE
    ) -> None:
        super().__init__(gwy, device_id)

        self._ref = gwy.get_device(device_id)  # device_factory(gwy, self._addr)
        self._tcs = self._ref.tcs

        self._change_counter: int = 8

        # initial values for the cycles, defer to reduce intial pkts Txs
        dt_now = dt.now()
        self._next_1f09_cycle: dt = dt_now + self._1F09_DURATION * 0.5
        self._next_3150_cycle: dt = dt_now + _3150_CYCLE_DURATION * 0.9
        self._next_3b00_cycle: dt = dt_now  # will defer below

        self._3b00_duration: td = td(seconds=0)  # not normally a valid duration
        self._cycle_rate: int = None  # type: ignore[assignment]
        self.cycle_rate = cycle_rate

        # Finally, 'start' the controller
        self._is_running = True
        self._tasks = [
            self._loop.create_task(self.tx_1f09_cycle()),
            self._loop.create_task(self.tx_3150_cycle()),
            self._loop.create_task(self.tx_3b00_cycle()),
        ]

    @property
    def cycle_rate(self) -> int:
        """Return the number of cycles per hour."""
        return self._cycle_rate

    @cycle_rate.setter
    def cycle_rate(self, value: int) -> None:
        """Set the number of cycles per hour.

        Calculates when the next cycle is due (it may become immediately due).
        """
        if value == self._cycle_rate:
            return
        assert value in (1, 2, 3, 6, 9, 12)
        self._cycle_rate = value
        self._3b00_duration, old_duration = (
            td(seconds=60 * 60 / value),
            self._3b00_duration,
        )
        self._next_3b00_cycle += self._3b00_duration - old_duration

    # @property
    # def heat_demand(self) -> int:
    #     """Return the number of cycles per hour."""
    #     return self._heat_demand

    @property
    def _sync_1f09_remaining(self) -> float:
        """Return the number of seconds until the next 1F09 cycle."""
        return (self._next_1f09_cycle - dt.now()).total_seconds()

    @property
    def _sync_3150_remaining(self) -> float:
        """Return the number of seconds until the next 3150 cycle."""
        return (self._next_3150_cycle - dt.now()).total_seconds()

    @property
    def _sync_3b00_remaining(self) -> float:
        """Return the number of seconds until the next 3B00 cycle."""
        return (self._next_3b00_cycle - dt.now()).total_seconds()

    def rx_frame_as_cmd(self, cmd: Command) -> None:
        """Find/Create an encoded frame, and place in on the ether."""

        LOOKUP_RQ: dict[str, Callable] = {
            Code._0005: self._make_0005,
            Code._0006: self._make_0006,
            Code._000C: self._make_000c,
            Code._1F09: self._make_1f09,
            # Code._30C9: self._make_30c9,
        }

        LOOKUP_OTH = {
            Code._0404: self._proc_0404,
            # Code._30C9: self._proc_30c9,
            Code._3150: self._proc_3150,
        }

        assert cmd.dst.id == self.id

        if cmd.code == Code._1FC9:  # TODO
            return

        if cmd.verb == RQ and (fnc := LOOKUP_RQ.get(cmd.code)):
            self.tx_frames_as_cmds(fnc(dest_id=cmd.src.id))

        elif cmd.verb != RQ and (fnc := LOOKUP_OTH.get(cmd.code)):
            self.tx_frames_as_cmds(fnc(cmd))

        elif response := RESPONSES.get(cmd.tx_header):
            self.tx_frames_as_cmds(self.make_response_pkt(response))

    async def tx_1f09_cycle(self) -> None:  # 1F09, 2309/30C9 or 000A
        """Periodically queue sync_cycle pkts on the ether for the gwy to Rx."""
        # I|1F09|FFxxxx and 2309/30C9 every 3-5m
        # I|1F09|FFxxxx and 000A every 60-90m, instead (say every 20 cycles?)

        # .I --- 01:087939 --:------ 01:087939 1F09 003 FF0532
        # .I --- 01:087939 --:------ 01:087939 2309 009 0007D0-010640-0201F4
        # .I --- 01:087939 --:------ 01:087939 30C9 009 0007A0-010634-020656

        counter = 0
        while self._is_running:
            await asyncio.sleep(self._sync_1f09_remaining)

            self._next_1f09_cycle = dt.now() + self._1F09_DURATION

            self.tx_frames_as_cmds(self._make_1f09())
            await asyncio.sleep(0.02)

            counter = (counter + 1) % _1F09_CYCLE_PERIOD
            if counter == 0:
                self.tx_frames_as_cmds(self._make_000a())
                return

            self.tx_frames_as_cmds(self._make_2309())
            await asyncio.sleep(0.02)
            self.tx_frames_as_cmds(self._make_30c9())

    async def tx_3150_cycle(self) -> None:  # 3150
        """Periodically queue heat_cycle pkts on the ether for the gwy to Rx."""
        # I|3150|FCxx minimum every cycle, and on state change & again 1m later

        while self._is_running:
            await asyncio.sleep(self._sync_3150_remaining)

            # if was an even-driven 3150 sent elsewhere, have to restart the timer
            if self._next_3150_cycle > (dt_now := dt.now()):
                continue
            else:
                self._next_3150_cycle = dt_now + _3150_CYCLE_DURATION

            self.tx_frames_as_cmds(self._make_3150())

    async def tx_3b00_cycle(self) -> None:  # 3B00
        """Periodically queue heat_cycle pkts on the ether for the gwy to Rx."""
        # I|3B00|FCC8 every 10m (usu. 3-4s drift)

        while self._is_running:
            await asyncio.sleep(self._sync_3b00_remaining)

            if self._sync_3b00_remaining > 0:  # may have been updated elsewhere
                continue

            self._next_3b00_cycle = dt.now() + self._3b00_duration

    def _make_0005(self, context: str) -> None | Command:
        def is_type(idx, zone_type):
            return (
                zones.get(f"{idx:02X}", {}).get(SZ_CLASS) == (ZON_ROLE_MAP[zone_type])
            )

        zone_type = context[2:]
        if not self._tcs or zone_type not in ZON_ROLE_MAP.HEAT_ZONES:
            return None

        zones = self._tcs.schema[SZ_ZONES]
        zone_mask = (is_type(idx, zone_type) for idx in range(0x10))

        return Command.put_zone_types(
            self.id, GWY_ID, zone_type, zone_mask, sub_idx=context[:2]
        )

    def _make_0006(self, dest_id: str) -> Command:
        payload = f"0005{self._change_counter:04X}"  # 0005 is not a typo
        return Command._from_attrs(
            RP, Code._0006, payload, addr0=self.id, addr1=dest_id
        )

    def _make_000a(self, dest_id: None | DeviceId = None) -> Command:  # TODO:
        return None  # type: ignore[return-value]

    def _make_000c(self, context: str) -> Command:  # TODO: no return None
        zone_idx, zone_type = context[:2], context[2:]

        if context == "000D":  # 01|DEV_ROLE_MAP.HTG
            pass
        elif context == "010D":  # 00|DEV_ROLE_MAP.HTG
            pass
        elif context == "000E":  # 00|DEV_ROLE_MAP.DHW
            pass
        elif context[2:] not in ZON_ROLE_MAP.HEAT_ZONES:
            raise NotImplementedError

        zone = self._tcs.zone_by_idx.get(zone_idx) if self._tcs else None
        if not zone:
            return None  # type: ignore[return-value]

        if zone_type == (SZ_ACTUATORS, zone.schema[SZ_CLASS]):
            pass
        if zone_type == (SZ_ACTUATORS, zone.schema[SZ_CLASS]):
            pass

        return None  # type: ignore[return-value]

    def _proc_0404(self, pkt: Command) -> Command:
        """Process an inbound 0404 packet."""
        if pkt.verb not in (W_, RQ):
            return None

        if response := RESPONSES.get(pkt.tx_header):
            if pkt.verb == W_:
                self._change_counter += 2
            return self.make_response_pkt(response)

    def _make_1100(self, dest_id: None | DeviceId = None) -> Command:  # TODO
        """Craft a stateful 1100 array pkt (is WIP)."""

        payload = "0007D00106400201F4"  # HACK
        return Command._from_attrs(
            I_, Code._1100, payload, addr0=self.id, addr2=self.id
        )

    def _make_1f09(self, dest_id: None | DeviceId = None) -> Command:
        """Craft a stateful 1F09 pkt."""

        if dest_id:
            verb = RP
            payload = f"00{self._sync_1f09_remaining * 10:04X}"
            addrs = {"addr0": self.id, "addr1": dest_id}
        else:
            # assert 0 < self._sync_cycle_remaining < 5  # required?
            verb = I_
            payload = f"FF{_1F09_CYCLE_DURATION * 10:04X}"  # TypeError: unsupported format string passed to datetime.timedelta.__format__
            addrs = {"addr0": self.id, "addr2": self.id}

        return Command._from_attrs(verb, Code._1F09, payload, **addrs)

    def _make_2309(
        self, dest_id: None | DeviceId = None, idx: None | _PktIdxT = None
    ) -> Command:  # WIP
        """Craft a stateful 2309 array pkt (is WIP)."""

        assert dest_id is None or idx is not None

        payload = "0007D00106400201F4"  # HACK
        return Command._from_attrs(
            I_, Code._2309, payload, addr0=self.id, addr2=self.id
        )

    def _make_30c9(
        self, dest_id: None | DeviceId = None, idx: None | _PktIdxT = None
    ) -> Command:  # WIP
        """Craft a stateful 3C09 array pkt (is WIP)."""

        assert dest_id is None or idx is not None

        payload = "0007A0010634020656"  # HACK
        return Command._from_attrs(
            I_, Code._30C9, payload, addr0=self.id, addr2=self.id
        )

    def _proc_30c9(self, cmd: Command) -> None:  # WIP
        """Process an inbound 30C9 packet."""
        pass

    def _make_3150(  # WIP
        self, dest_id: None | DeviceId = None, idx: None | _PktIdxT = None
    ) -> Command:
        """Craft a stateful 3150 cmd (is WIP)."""

        payload = "0007A0010634020656"  # HACK
        return Command._from_attrs(
            I_, Code._30C9, payload, addr0=self.id, addr2=self.id
        )

    def _proc_3150(self, cmd: Command) -> None:  # WIP
        """Process an inbound 3150 packet."""
        pass

    def _make_3b00(self) -> Command:
        """Craft a 3B00 pkt."""
        # 'I --- 01:145038 --:------ 01:145038 3B00 002 FCC8  # only periodic I, never RP
        # .I --- --:------ --:------ 12:207082 3B00 002 00C8  # idx=00, never? FC
        # .I --- 13:209679 --:------ 13:209679 3B00 002 00C8  #

        # 2020-06-09T12:06:45.628934 054  I --- 13:209679 --:------ 13:209679 3B00 002 00C8
        # 2020-06-09T12:06:48.415689 045  I --- 01:158182 --:------ 01:158182 3B00 002 FCC8
        # 2020-06-09T12:16:44.544482 057  I --- 13:209679 --:------ 13:209679 3B00 002 00C8
        # 2020-06-09T12:16:47.107356 045  I --- 01:158182 --:------ 01:158182 3B00 002 FCC8
        # does 13: Tx at end of cycle?

        return Command._from_attrs(I_, Code._30C9, "FCC8", addr0=self.id, addr2=self.id)

    def add_zone(
        self,
        name: ZoneNameT,
        heating_type: ZoneTypeT,
        use_evotouch: bool = False,
        **params: ZoneParamsT,
    ) -> ZoneIdxT:
        """Add a zone."""

        zone_idx: ZoneIdxT = f"{[i for i in range(12) if i not in self._zones][0]:02X}"
        self._zones[zone_idx] = MockZone(
            self, zone_idx, name, heating_type, use_evotouch, **params
        )

        # send 0005

        # self.bind_zone_sensor(zone_idx, use_evotouch=use_evotouch)
        # self.bind_zone_actuator(zone_idx)
        return zone_idx

    def bind_zone_sensor(self, zone_idx: ZoneIdxT, use_evotouch: bool = None) -> None:
        """Bind a sensor to a zone.

        Clear the old sensor, if there was one.
        """

        def callback(msg, *args, **kwargs) -> None:
            """If the bind was successful, set the zone sensor."""
            if msg:
                zone._set_sensor(msg.src.id)

        zone: MockZone = self._zones[zone_idx]

        if use_evotouch is True:
            zone._set_sensor(self._ref.id)
        elif use_evotouch is False or use_evotouch is None:
            zone._set_sensor(None)
            self._ref._bind_waiting("30C9", idx=zone_idx, callback=callback)
        else:
            raise TypeError(
                f"Invalid {SZ_USE_EVOTOUCH} parameter for zone: {use_evotouch}"
            )

        # send 000C

    def bind_zone_actuator(self, zone_idx: ZoneIdxT):
        """Bind an actuator a zone.

        Don't clear any previous actuators.
        """
        pass

    def configure_zone(
        self,
        zone_idx: ZoneIdxT,
        name: ZoneNameT = None,
        heating_type: ZoneTypeT = None,
        use_evotouch: bool = False,
        **params: ZoneParamsT,
    ):
        """Configure a zone."""
        zone: MockZone = self._zone[zone_idx]

        if name is not None:
            zone.name = name
        if heating_type is not None:
            zone.heating_type = heating_type
        if use_evotouch is not None:
            self.bind_zone_sensor(zone_idx, use_evotouch=use_evotouch)
        if params:
            zone.parameters = params

    def del_zone(self, zone_idx: ZoneIdxT):
        """Delete a zone."""
        del self._zones[zone_idx]

    def factory_reset(self):
        """Perform a factory reset."""
        self._appliance_control = None
        self._dhw_zone = None
        self._zones = []

        # send 0005, etc.

    def bind_appliance_control(self, appliance_type: str = "WRB"):  # to check UI
        """Bind the appliance controller: WRB (BDR), OTB, or HPR (BDR91T)."""
        pass

    def add_dhw(self):  # to check UI
        """Add the DHW zone."""
        self._dhw_zone = None

    def bind_dhw_sensor(self):  # to check UI
        """Bind a sensor to the DHW zone."""
        pass

    def del_dhw(self):  # to check UI
        """Delete the DHW zone."""
        self._dhw_zone = None


class MockDeviceThm(MockDeviceBase):
    """A pseudo-mocked thermostat used for testing."""

    def __init__(self, gwy, device_id, *, schema=None) -> None:
        super().__init__(gwy, device_id)

        self.temperature = None  # TODO: maintain internal state (is needed?)


RESPONSES: dict[str, str] = {
    f"0006|RQ|{CTL_ID}": f"RP --- {CTL_ID} {GWY_ID} --:------ 0006 004 00050008",
    # RQ schedule for zone 01
    f"0404|RQ|{CTL_ID}|0101": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 048 0120000829010368816DCCC91183301005D1D93428200E1C7D720C04402C0442640E82000C851701ADD3AFAED1131151",
    f"0404|RQ|{CTL_ID}|0102": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 048 0120000829020339DEBC8DBE1EFBDB5EDBA8DDB92DBEDFADDAB6671179E4FF4EC153F0143C05CFC033F00C3C03CFC173",
    f"0404|RQ|{CTL_ID}|0103": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 046 01200008270303F01C3C072FC00BF002BC00AF7CFEB6DEDE46BBB721EE6DBA78095E8297E0E5CF5BF50DA0291B9C",
    # RQ schedule for DHW
    f"0404|RQ|{CTL_ID}|HW01": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 048 0023000829010468816DCDD10980300C45D1BE24CD9713398093388BF33981A3882842B5BDE9571F178E4ABB4DA5E879",
    f"0404|RQ|{CTL_ID}|HW02": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 048 00230008290204EDCEF7F3DD761BBBC9C7EEF0B15BEABF13B80257E00A5C812B700D5C03D7C035700D5C03D7C175701D",
    f"0404|RQ|{CTL_ID}|HW03": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 048 002300082903045C07D7C1757003DC0037C00D7003DC00B7827B6FB38D5DEF56702BB8F7B6766E829BE026B8096E829B",
    f"0404|RQ|{CTL_ID}|HW04": f"RP --- {CTL_ID} {GWY_ID} --:------ 0404 014 002300080704041FF702BAC2188E",
    # W schedule for zone 01
    f"0404| W|{CTL_ID}|0101": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 01200008290103",
    f"0404| W|{CTL_ID}|0102": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 01200008290203",
    f"0404| W|{CTL_ID}|0103": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 01200008270300",
    # W schedule for DHW
    f"0404| W|{CTL_ID}|HW01": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 00230008290104",
    f"0404| W|{CTL_ID}|HW02": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 00230008290204",
    f"0404| W|{CTL_ID}|HW03": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 00230008290304",
    f"0404| W|{CTL_ID}|HW04": f" I --- {CTL_ID} {GWY_ID} --:------ 0404 007 00230008070400",
    #
    f"2309|RQ|{CTL_ID}|00": f"RP --- {CTL_ID} --:------ {GWY_ID} 2309 003 0007D0",
    f"2309|RQ|{CTL_ID}|01": f"RP --- {CTL_ID} --:------ {GWY_ID} 2309 003 010640",
    f"2309|RQ|{CTL_ID}|02": f"RP --- {CTL_ID} --:------ {GWY_ID} 2309 003 0201F4",
    #
    f"30C9|RQ|{CTL_ID}|00": f"RP --- {CTL_ID} --:------ {GWY_ID} 30C9 003 0007A0",
    f"30C9|RQ|{CTL_ID}|01": f"RP --- {CTL_ID} --:------ {GWY_ID} 30C9 003 010634",
    f"30C9|RQ|{CTL_ID}|02": f"RP --- {CTL_ID} --:------ {GWY_ID} 30C9 003 020656",
}  # "pkt_header": "response_pkt"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - RAMSES-II compatible packet transport.

Operates at the pkt layer of: app - msg - pkt - h/w

For ser2net, use the following YAML with: ser2net -c hgi80.yaml
  connection: &con00
  accepter: telnet(rfc2217),tcp,5001
  timeout: 0
  connector: serialdev,/dev/ttyUSB0,115200n81,local
  options:
    max-connections: 3

For socat, see:
  socat -dd pty,raw,echo=0 pty,raw,echo=0
  python client.py monitor /dev/pts/0
  cat packet.log | cut -d ' ' -f 2- | unix2dos > /dev/pts/1

For re-flashing evofw3 via Arduino IDE on *my* atmega328p (YMMV):
 - Board:      atmega328p (SW UART)
 - Bootloader: Old Bootloader
 - Processor:  atmega328p (5V, 16 MHz)
 - Host:       57600 (or 115200, YMMV)
 - Pinout:     Nano

For re-flashing evofw3 via Arduino IDE on *my* atmega32u4 (YMMV):
 - Board:      atmega32u4 (HW UART)
 - Processor:  atmega32u4 (5V, 16 MHz)
 - Pinout:     Pro Micro
"""

# TODO:
# - add auto-detection of evofw3/HGI80
# - chase down gwy.config.disable_discovery
# - chase down / check deprecation


from __future__ import annotations

import asyncio
import functools
import logging
import os
import re
from collections.abc import Callable, Iterable
from datetime import datetime as dt
from io import TextIOWrapper
from string import printable
from typing import TYPE_CHECKING, Any

import serial_asyncio  # type: ignore[import-untyped]
from serial import (  # type: ignore[import-untyped]
    Serial,
    SerialException,
    serial_for_url,
)
from serial.tools.list_ports import comports  # type: ignore[import-untyped]

from .address import NON_DEV_ADDR, NUL_DEV_ADDR
from .command import Command
from .const import (
    DEV_TYPE_MAP,
    SZ_ACTIVE_HGI,
    SZ_IS_EVOFW3,
    SZ_KNOWN_HGI,
    SZ_SIGNATURE,
    DevType,
    __dev_mode__,
)
from .exceptions import PacketInvalid, TransportSerialError, TransportSourceInvalid
from .helpers import dt_now
from .packet import Packet
from .schemas import (
    SCH_SERIAL_PORT_CONFIG,
    SZ_BLOCK_LIST,
    SZ_CLASS,
    SZ_DISABLE_QOS,
    SZ_DISABLE_SENDING,
    SZ_EVOFW_FLAG,
    SZ_INBOUND,
    SZ_KNOWN_LIST,
    SZ_OUTBOUND,
)

# skipcq: PY-W2000
from .const import (  # noqa: F401, isort: skip, pylint: disable=unused-import
    I_,
    RP,
    RQ,
    W_,
    Code,
)

if TYPE_CHECKING:  # mypy TypeVars and similar (e.g. Index, Verb)
    # skipcq: PY-W2000
    from .address import DeviceId
    from .const import Index, Verb  # noqa: F401, pylint: disable=unused-import


if TYPE_CHECKING:
    from . import QosProtocol as _ProtocolT

_SIGNATURE_MAX_TRYS = 24
_SIGNATURE_GAP_SECS = 0.05

TIP = f", configure the {SZ_KNOWN_LIST}/{SZ_BLOCK_LIST} as required"


DEV_MODE = __dev_mode__ and False

_LOGGER = logging.getLogger(__name__)
# _LOGGER.setLevel(logging.WARNING)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)

# All debug flags should be False for end-users
_DEBUG_DISABLE_REGEX_WARNINGS = True  # useful for dev/test
_DEBUG_FORCE_LOG_FRAMES = False  # useful for dev/test


def _normalise(pkt_line: str) -> str:
    """Perform any (transparent) frame-level hacks, as required at (near-)RF layer.

    Goals:
    - ensure an evofw3 provides the same output as a HGI80 (none, presently)
    - handle 'strange' packets (e.g. I/08:/0008)
    """

    # psuedo-RAMSES-II packets...
    if pkt_line[10:14] in (" 08:", " 31:") and pkt_line[-16:] == "* Checksum error":
        pkt_line = pkt_line[:-17] + " # Checksum error (ignored)"

    return pkt_line.strip()


def _str(value: bytes) -> str:
    try:
        result = "".join(
            c
            for c in value.decode("ascii", errors="strict")  # was: .strip()
            if c in printable
        )
    except UnicodeDecodeError:
        _LOGGER.warning("%s < Cant decode bytestream (ignoring)", value)
        return ""
    return result


class _PktMixin:
    """Base class for RAMSES II transports."""

    _this_pkt: None | Packet
    _prev_pkt: None | Packet
    _protocol: _ProtocolT

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._this_pkt: Packet = None
        self._prev_pkt: Packet = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._protocol})"

    def _pkt_received(self, pkt: Packet) -> None:
        """Pass any valid/wanted Packets to the protocol's callback.

        Also maintain _prev_pkt, _this_pkt attrs.
        """

        self._this_pkt, self._prev_pkt = pkt, self._this_pkt

        # NOTE: No need to use call_soon() here, and they may break Qos/Callbacks
        # NOTE: Thus, excepts need checking
        try:  # below could be a call_soon?
            self._protocol.pkt_received(pkt)
        except AssertionError as exc:  # protect from upper-layer callbacks
            _LOGGER.exception("%s < exception from msg layer: %s", pkt, exc)


class _DeviceIdFilterMixin:  # NOTE: active gwy detection in here too
    """Filter out any unwanted (but otherwise valid) packets via device ids."""

    _extra: dict[str, Any]  # mypy

    def __init__(
        self,
        *args,
        enforce_include_list: bool = False,
        exclude_list: None | dict = None,
        include_list: None | dict = None,
        **kwargs,
    ) -> None:
        exclude_list = exclude_list or {}
        include_list = include_list or {}

        self._evofw_flag = kwargs.pop(SZ_EVOFW_FLAG, None)  # gwy.config.evofw_flag

        super().__init__(*args, **kwargs)

        self.enforce_include = enforce_include_list
        self._exclude = list(exclude_list.keys())
        self._include = list(include_list.keys()) + [NON_DEV_ADDR.id, NUL_DEV_ADDR.id]

        self._unwanted: list = []  # not: [NON_DEV_ADDR.id, NUL_DEV_ADDR.id]

        for key in (SZ_ACTIVE_HGI, SZ_SIGNATURE, SZ_KNOWN_HGI):
            self._extra[key] = None

        known_hgis = [
            k for k, v in include_list.items() if v.get(SZ_CLASS) == DevType.HGI
        ]
        if not known_hgis:
            _LOGGER.warning(
                f"The {SZ_KNOWN_LIST} should include the gateway (HGI) but doesn't{TIP}"
            )
        else:
            self._extra[SZ_KNOWN_HGI] = known_hgis[0]
        if len(known_hgis) > 1:
            _LOGGER.warning(
                f"The {SZ_KNOWN_LIST} should have only 1 gateway (HGI) but has more"
                f" (the selected Known gateway is: {self._extra[SZ_KNOWN_HGI]}){TIP}"
            )

    def _set_active_hgi(self, dev_id: DeviceId, by_signature: bool = False) -> None:
        """Set the Active Gateway device (HGI), warn if it is filtered incorrectly."""
        msg = "Active gateway " + "(by signature)" if by_signature else "(by filter)"

        if dev_id in self._exclude:
            _LOGGER.error(f"{msg} is in {SZ_BLOCK_LIST}: {dev_id}{TIP}")
        elif dev_id in self._include:
            _LOGGER.info(f"{msg}: {dev_id}")
        else:
            _LOGGER.warning(f"{msg} not in {SZ_KNOWN_LIST}: {dev_id}{TIP}")
        self._extra[SZ_ACTIVE_HGI] = dev_id

    def _is_wanted_addrs(
        self, src_id: DeviceId, dst_id: DeviceId, payload: None | str = None
    ) -> bool:
        """Return True if the packet is not to be filtered out.

        In any one packet, an excluded device_id 'trumps' an included device_id.
        """

        def deprecate_foreign_hgi(dev_id: DeviceId) -> None:
            self._unwanted.append(dev_id)
            _LOGGER.warning(
                f"Blacklisted a Foreign gateway (is it a HVAC device?): {dev_id}"
                f" (Active gateway is: {self._extra[SZ_ACTIVE_HGI]}){TIP}"
            )

        for dev_id in dict.fromkeys((src_id, dst_id)):  # removes duplicates
            # TODO: _unwanted exists since (in future) stale entries need to be removed

            if dev_id in self._exclude or dev_id in self._unwanted:
                return False

            if dev_id == self._extra[SZ_ACTIVE_HGI]:  # is active gwy
                continue

            if dev_id in self._include:  # incl. 63:262142 & --:------
                continue

            if self.enforce_include:
                return False

            if dev_id[:2] != DEV_TYPE_MAP.HGI:
                continue

            if self._extra[SZ_ACTIVE_HGI]:
                deprecate_foreign_hgi(dev_id)  # self._unwanted.append(dev_id)
                return False

            if dev_id == src_id and payload == self._extra[SZ_SIGNATURE]:
                self._set_active_hgi(dev_id)
                continue

            if dev_id == self._extra[SZ_KNOWN_HGI]:
                self._set_active_hgi(dev_id)

        return True

    def _pkt_received(self, pkt: Packet) -> None:
        """Validate a Packet and dispatch it to the protocol's callback."""
        if self._is_wanted_addrs(pkt.src.id, pkt.dst.id, pkt.payload):
            super()._pkt_received(pkt)  # type: ignore[misc]


class _RegHackMixin:
    def __init__(self, *args, use_regex: None | dict = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        use_regex = use_regex or {}

        self.__inbound_rule = use_regex.get(SZ_INBOUND, {})
        self.__outbound_rule = use_regex.get(SZ_OUTBOUND, {})

    @staticmethod
    def __regex_hack(pkt_line: str, regex_rules: dict) -> str:
        if not regex_rules:
            return pkt_line

        result = pkt_line
        for k, v in regex_rules.items():
            try:
                result = re.sub(k, v, result)
            except re.error as exc:
                _LOGGER.warning(f"{pkt_line} < issue with regex ({k}, {v}): {exc}")

        if result != pkt_line and not _DEBUG_DISABLE_REGEX_WARNINGS:
            (_LOGGER.debug if DEV_MODE else _LOGGER.warning)(
                f"{pkt_line} < Changed by use_regex to: {result}"
            )
        return result

    def _frame_received(self, dtm: str, frame: str) -> None:
        super()._frame_received(dtm, self.__regex_hack(frame, self.__inbound_rule))  # type: ignore[misc]

    def _send_frame(self, frame: str) -> None:
        super()._send_frame(self.__regex_hack(frame, self.__outbound_rule))  # type: ignore[misc]


class _FileTransport(_PktMixin, asyncio.ReadTransport):
    """Parse a file (or a dict) for packets, and never send."""

    READER_TASK = "reader_task"
    _protocol: _ProtocolT

    _dtm_str: str = None  # type: ignore[assignment]  # FIXME: remove this somehow

    def __init__(
        self,
        protocol: _ProtocolT,
        pkt_source: dict | TextIOWrapper,
        loop: None | asyncio.AbstractEventLoop = None,
        extra: None | dict = None,
    ) -> None:
        super().__init__(extra=extra)

        self._pkt_source = pkt_source
        self._protocol = protocol
        self._loop: asyncio.AbstractEventLoop = loop or asyncio.get_running_loop()

        self._is_closing: bool = False
        self._is_reading: bool = False

        self._reader_task = self._loop.create_task(self._start_reader())

    def _dt_now(self) -> dt:
        """Return a precise datetime, using a packet's dtm field."""

        try:
            return dt.fromisoformat(self._dtm_str)  # always current pkt's dtm
        except (TypeError, ValueError):
            pass

        try:
            return self._this_pkt.dtm  # type: ignore[union-attr]
        except AttributeError:
            return dt(1970, 1, 1, 1, 0)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """The asyncio event loop as used by SerialTransport."""
        return self._loop

    def get_extra_info(self, name, default=None) -> Any:
        if name == self.READER_TASK:
            return self._reader_task
        return super().get_extra_info(name, default)

    def is_closing(self) -> bool:
        """Return True if the transport is closing or closed."""
        return self._is_closing

    def is_reading(self):
        """Return True if the transport is receiving."""
        return self._is_reading

    def pause_reading(self) -> None:
        """Pause the receiving end (no data to protocol.pkt_received())."""
        self._is_reading = False

    def resume_reading(self) -> None:
        """Resume the receiving end."""
        self._is_reading = True

    async def _start_reader(self) -> None:  # TODO
        self._is_reading = True
        try:
            await self._reader()
        except KeyboardInterrupt as exc:
            self._protocol.connection_lost(exc)
        else:
            self._protocol.connection_lost(None)

    async def _reader(self) -> None:  # TODO
        """Loop through the packet source for Frames and process them."""

        if isinstance(self._pkt_source, dict):
            for dtm_str, pkt_line in self._pkt_source.items():  # assume dtm_str is OK
                while not self._is_reading:
                    await asyncio.sleep(0.001)
                self._frame_received(dtm_str, pkt_line)
                await asyncio.sleep(0)  # NOTE: big performance penalty if delay >0

        elif isinstance(self._pkt_source, TextIOWrapper):
            for dtm_pkt_line in self._pkt_source:  # should check dtm_str is OK
                while not self._is_reading:
                    await asyncio.sleep(0.001)
                self._frame_received(dtm_pkt_line[:26], dtm_pkt_line[27:])
                await asyncio.sleep(0)  # NOTE: big performance penalty if delay >0

        else:
            raise TransportSourceInvalid(
                f"Packet source is not dict or TextIOWrapper: {self._pkt_source:!r}"
            )

    def _frame_received(self, dtm_str: str, frame: str) -> None:
        """Make a Packet from the Frame and process it."""
        self._dtm_str = dtm_str  # HACK: FIXME: remove need for this, somehow

        try:
            pkt = Packet.from_file(dtm_str, frame)  # is OK for when src is dict
        except (PacketInvalid, ValueError):  # VE from dt.fromisoformat()
            return
        self._pkt_received(pkt)

    def close(self, exc: None | Exception = None) -> None:
        """Close the transport (calls self._protocol.connection_lost())."""
        if self._is_closing:
            return
        self._is_closing = True

        if self._reader_task:
            self._reader_task.cancel()

        self._loop.call_soon(self._protocol.connection_lost, exc)


class _PortTransport(_PktMixin, serial_asyncio.SerialTransport):
    """Poll a serial port for packets, and send (without QoS)."""

    loop: asyncio.AbstractEventLoop
    serial: Serial

    _init_fut: asyncio.Future
    _init_task: None | asyncio.Task = None

    _recv_buffer: bytes = b""

    def __init__(
        self,
        protocol: _ProtocolT,
        pkt_source: Serial,
        loop: None | asyncio.AbstractEventLoop = None,
        extra: None | dict = None,
    ) -> None:
        super().__init__(loop or asyncio.get_running_loop(), protocol, pkt_source)

        self._extra: dict = {} if extra is None else extra

    def _dt_now(self) -> dt:
        """Return a precise datetime, using the curent dtm."""
        return dt_now()

    def _read_ready(self) -> None:
        # data to self._bytes_received() instead of self._protocol.data_received()
        try:
            data: bytes = self._serial.read(self._max_read_size)
        except SerialException as e:
            self._close(exc=e)
            return

        if data:
            self._bytes_received(data)  # was: self._protocol.pkt_received(data)

    def is_reading(self) -> None:
        """Return True if the transport is receiving."""
        return self._has_reader

    def _bytes_received(self, data: bytes) -> None:  # logs: RCVD(bytes)
        """Make a Frame from the data and process it."""

        def bytes_received(data: bytes) -> Iterable[tuple[dt, bytes]]:
            self._recv_buffer += data
            if b"\r\n" in self._recv_buffer:
                lines = self._recv_buffer.split(b"\r\n")
                self._recv_buffer = lines[-1]
                for line in lines[:-1]:
                    yield self._dt_now(), line + b"\r\n"

        for dtm, raw_line in bytes_received(data):
            if _DEBUG_FORCE_LOG_FRAMES:
                _LOGGER.warning("Rx: %s", raw_line)
            elif _LOGGER.getEffectiveLevel() == logging.INFO:  # log for INFO not DEBUG
                _LOGGER.info("Rx: %s", raw_line)
            self._frame_received(dtm, _normalise(_str(raw_line)))

    def _frame_received(self, dtm: dt, frame: str) -> None:
        """Make a Packet from the Frame and process it."""

        try:
            pkt = Packet.from_port(dtm, frame)
        except (PacketInvalid, ValueError):  # VE from dt.fromisoformat()
            return

        if (
            not self._init_fut.done()
            and pkt.code == Code._PUZZ
            and pkt.payload == self._extra[SZ_SIGNATURE]
        ):
            self._set_active_hgi(pkt.src.id, by_signature=True)
            self._init_fut.set_result(pkt)

        self._pkt_received(pkt)  # TODO: remove raw_line attr from Packet()

    def send_frame(self, frame: str) -> None:  # Protocol usu. calls this, not write()
        self._send_frame(frame)

    def _send_frame(self, frame: str) -> None:
        self.write(bytes(frame, "ascii") + b"\r\n")

    def write(self, data: bytes) -> None:  # logs: SENT(bytes)
        if _DEBUG_FORCE_LOG_FRAMES:
            _LOGGER.warning("Tx:     %s", data)
        elif _LOGGER.getEffectiveLevel() == logging.INFO:  # log for INFO not DEBUG
            _LOGGER.info("Tx:     %s", data)
        super().write(data)

    def close(self, exc: None | Exception = None) -> None:
        """Close the transport (calls self._protocol.connection_lost())."""
        super().close()
        if self._init_task:
            self._init_task.cancel()


# ### Read-Only Transports for dict / log file ########################################
class FileTransport(_DeviceIdFilterMixin, _FileTransport):
    """Parse a file (or a dict) for packets, and never send."""

    def __init__(self, *args, disable_sending: bool = True, **kwargs) -> None:
        if disable_sending is False:
            raise TransportSourceInvalid("This Transport cannot send packets")
        super().__init__(*args, **kwargs)
        self.loop.call_soon(self._protocol.connection_made, self)


# ### Read-Write Transport for serial port ############################################
class PortTransport(_RegHackMixin, _DeviceIdFilterMixin, _PortTransport):
    """Poll a serial port for packets, and send (without QoS)."""

    _init_fut: asyncio.Future
    _init_task: asyncio.Task

    def __init__(self, *args, disable_sending: bool = False, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._is_hgi80 = self.is_hgi80(self.serial.name)
        self._make_connection(disable_sending)

    def _make_connection(self, sending_disabled: bool) -> None:
        """Call connection_made() after housekeeping functions are completed."""

        # HGI80s (and also VMs) take longer to send signature packets as they have long
        # initialisation times, so we must wait until they send OK

        # signature also serves to discover the HGI's device_id (& for pkt log, if any)

        # Could instead have: connection_made(self, pkt=pkt) *if* pkt is sig. echo, but
        # would require a re-write or portions of both Transport & Protocol

        def call_make_connection() -> None:
            """Invoke the Protocol.connection_made() callback."""
            # if self._is_hgi80 is not True:  # TODO: !V doesn't work, why?
            #     self._send_frame("!V")

            self.loop.call_soon(
                functools.partial(self._protocol.connection_made, self, ramses=True)
            )  # was: self._protocol.connection_made(self, ramses=True)

        async def connect_without_signature() -> None:
            """Call connection_made() without sending/waiting for a signature."""
            self._init_fut.set_result(None)
            call_make_connection()

        async def connect_after_signature() -> None:
            """Poll port with signatures, call connection_made() after first echo."""
            sig = Command._puzzle()
            self._extra[SZ_SIGNATURE] = sig.payload

            num_sends = 0
            while num_sends < _SIGNATURE_MAX_TRYS:
                num_sends += 1

                self._send_frame(str(sig))
                await asyncio.sleep(_SIGNATURE_GAP_SECS)

                if self._init_fut.done():
                    call_make_connection()
                    return

            self._init_fut.set_exception(
                TransportSerialError("Never received an echo signature")
            )

        self._init_fut = asyncio.Future()
        if sending_disabled:
            self._init_task = asyncio.create_task(connect_without_signature())
        else:  # incl. disable_qos
            self._init_task = asyncio.create_task(connect_after_signature())

    @staticmethod
    def is_hgi80(serial_port: SerPortName) -> None | bool:
        """Return True/False if the device attached to the port is/isn't an HGI80.

        Return None if it's not possible to tell (effectively assume is evofw3).
        """

        vid = {x.device: x.vid for x in comports()}.get(serial_port)

        if vid and vid == 0x10AC:  # aka Honeywell, Inc.
            _LOGGER.debug(f"{serial_port}: is HGI80-compatible (by VID)")
            return True

        product: None | str = {
            x.device: getattr(x, "product", None) for x in comports()
        }.get(serial_port)

        if not product:  # is None - not member of plugdev group?
            pass
        # elif "TUSB3410" in product:  # ?needed
        #     _LOGGER.info("The gateway is HGI80-compatible (by USB attrs)")
        #     return True
        elif "evofw3" in product or "FT232R" in product:
            _LOGGER.debug(f"{serial_port}: appears evofw3-compatible (by USB attrs)")
            return False

        _LOGGER.warning(
            f"{serial_port}: the gateway type is not determinable, will assume evofw3 "
            "(check you have the rights to enumerate USB attrs?)"
        )
        return None  # try sending an "!V", expect "# evofw3 0.7.1"

    def get_extra_info(self, name, default=None):
        if name == SZ_IS_EVOFW3:
            return not self._is_hgi80  # NOTE: None (unknown) as False (is_evofw3)
        return self._extra.get(name, default)


# ### Read-Write Transport *with QoS* for serial port #################################
class QosTransport(PortTransport):
    """Poll a serial port for packets, and send with QoS."""

    # NOTE: Might normally include code to the limit duty cycle, etc., but see
    # the note in Protocol layer

    pass


def find_gateway_device() -> None | SerPortName:
    """Find the gateway device and return its port name (assumes exactly one)."""
    port_names = [
        p.device for p in comports() if PortTransport.is_hgi80(p.device) is not None
    ]
    try:
        return port_names[0]
    except IndexError:
        return None


async def transport_factory(
    protocol: Callable[[], _ProtocolT],
    /,
    *,
    port_name: None | SerPortName = None,
    port_config: None | dict = None,
    packet_log: None | TextIOWrapper = None,
    packet_dict: None | dict = None,
    **kwargs,
) -> RamsesTransportT:
    """Create and return a Ramses-specific async packet Transport."""

    # The kwargs must be a subset of: loop, extra, and...
    # disable_sending, enforce_include_list, exclude_list, include_list, use_regex

    async def poll_until_connection_made(protocol: _ProtocolT):
        """Poll until the Transport is bound to the Protocol."""
        while protocol._transport is None:
            await asyncio.sleep(0.005)

    def get_serial_instance(ser_name: SerPortName, ser_config: dict) -> Serial:
        # For example:
        # - python client.py monitor 'rfc2217://localhost:5001'
        # - python client.py monitor 'alt:///dev/ttyUSB0?class=PosixPollSerial'

        ser_config = SCH_SERIAL_PORT_CONFIG(ser_config or {})

        try:
            ser_obj = serial_for_url(ser_name, **ser_config)
        except SerialException as exc:
            _LOGGER.error(
                "Failed to open %s (config: %s): %s", ser_name, ser_config, exc
            )
            raise TransportSerialError(
                f"Unable to open the serial port: {ser_name}"
            ) from exc

        # FTDI on Posix/Linux would be a common environment for this library...
        try:
            ser_obj.set_low_latency_mode(True)
        except (AttributeError, NotImplementedError, ValueError):
            pass  # Wrong OS/Platform/not FTDI

        return ser_obj

    def issue_warning() -> None:
        """Warn of the perils of semi-supported configurations."""
        _LOGGER.warning(
            f"{'Windows' if os.name == 'nt' else 'This type of serial interface'} "
            "is not fully supported by this library: "
            "please don't report any Transport/Protocol errors/warnings, "
            "unless they are reproducable with a standard configuration "
            "(e.g. linux with a local serial port)"
        )

    if len([x for x in (packet_dict, packet_log, port_name) if x is not None]) != 1:
        raise TransportSourceInvalid(
            "Packet source must be exactly one of: packet_dict, packet_log, port_name"
        )

    if (pkt_source := packet_log or packet_dict) is not None:
        return FileTransport(protocol, pkt_source, **kwargs)

    assert port_name is not None  # mypy
    assert port_config is not None  # mypy

    # may: raise TransportSerialError("Unable to open serial port...")
    ser_instance = get_serial_instance(port_name, port_config)

    # TODO: test these...
    if os.name == "nt" or ser_instance.portstr[:7] in ("rfc2217", "socket:"):
        issue_warning()
        # return PortTransport(protocol, ser_instance, **kwargs)

    if kwargs.get(SZ_DISABLE_SENDING) or kwargs.get(SZ_DISABLE_QOS):  # no need for QoS
        transport = PortTransport(protocol, ser_instance, **kwargs)
    else:
        transport = QosTransport(protocol, ser_instance, **kwargs)

    # wait to get (first) signature echo from evofw3/HGI80 (even if disable_sending)
    try:
        await asyncio.wait_for(transport._init_fut, timeout=3)  # signature echo
    except asyncio.TimeoutError as exc:
        raise TransportSerialError("Transport did not initialise successfully") from exc

    # wait for protocol to receive connection_made(transport) (i.e. is quiesced)
    try:
        await asyncio.wait_for(poll_until_connection_made(protocol), timeout=3)
    except asyncio.TimeoutError as exc:
        raise TransportSerialError("Transport did not bind to Protocol") from exc

    return transport


RamsesTransportT = FileTransport | PortTransport | QosTransport
SerPortName = str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Test the various helper APIs.
"""

# TODO: add test for ramses_tx.frame.pkt_header()

from ramses_rf.const import DEV_ROLE_MAP, DEV_TYPE_MAP
from ramses_rf.helpers import merge
from ramses_tx.const import attr_dict_factory
from tests.helpers import assert_raises


def test_merge_dicts() -> None:
    """Deep merge a src dict (precident) into a dst dict and return the result."""

    src = {"top": {"deep": {"in_both": "1", "in_src": "dog"}}}
    dst = {"top": {"deep": {"in_both": "9", "in_dst": "cat"}}}
    out = {"top": {"deep": {"in_both": "1", "in_src": "dog", "in_dst": "cat"}}}
    assert out == merge(src, dst)

    assert out != dst
    assert out == merge(src, dst, _dc=True)
    assert out == dst

    src = {"top": {"deep": {"in_both": [0, 1]}}}
    dst = {"top": {"deep": {"in_both": [0, 9]}}}
    out = {"top": {"deep": {"in_both": [0, 1, 9]}}}
    assert out == merge(src, dst)

    src = {"top": {"deep": {"in_both": "non-list"}}}
    dst = {"top": {"deep": {"in_both": [0, 9]}}}
    out = {"top": {"deep": {"in_both": "non-list"}}}
    assert out == merge(src, dst)

    src = {"top": {"deep": {"in_both": [0, 1]}}}
    dst = {"top": {"deep": {"in_both": "non-list"}}}
    out = {"top": {"deep": {"in_both": [0, 1]}}}
    assert out == merge(src, dst)


def test_attrdict_class() -> None:
    _ = attr_dict_factory(MAIN_DICT, attr_table=ATTR_DICT)

    assert_raises(KeyError, DEV_TYPE_MAP.slug, "_rubbish_")
    assert_raises(KeyError, DEV_TYPE_MAP.slug, None)
    # assert DEV_TYPE_MAP.slug(None), "DEV")

    try:
        DEV_ROLE_MAP["_08"]
    except KeyError:
        pass
    else:
        assert False

    assert DEV_ROLE_MAP.DHW == "0D"
    assert DEV_ROLE_MAP._0D == "dhw_sensor"
    assert DEV_ROLE_MAP.DHW_SENSOR == "0D"

    assert DEV_ROLE_MAP["RAD"] == "rad_actuator"
    assert DEV_ROLE_MAP["08"] == "rad_actuator"
    assert DEV_ROLE_MAP["rad_actuator"] == "08"

    assert DEV_ROLE_MAP._hex("SEN") == "04"
    assert_raises(KeyError, DEV_ROLE_MAP._hex, "04")  # aka: DEV_ROLE_MAP._hex("04")
    assert DEV_ROLE_MAP._hex("zone_sensor") == "04"

    assert_raises(KeyError, DEV_ROLE_MAP._hex, "_rubbish_")
    assert_raises(KeyError, DEV_ROLE_MAP._hex, None)

    assert DEV_ROLE_MAP._str("OUT") == "out_sensor"
    assert DEV_ROLE_MAP._str("0C") == "out_sensor"
    assert_raises(KeyError, DEV_ROLE_MAP._str, "out_sensor")

    assert_raises(KeyError, DEV_ROLE_MAP._str, "_rubbish_")
    assert_raises(KeyError, DEV_ROLE_MAP._str, None)

    assert_raises(KeyError, DEV_ROLE_MAP.slug, "RFG")  # aka: DEV_ROLE_MAP.slug("RFG")
    assert DEV_ROLE_MAP.slug("10") == "RFG"
    assert DEV_ROLE_MAP.slug("remote_gateway") == "RFG"

    assert_raises(KeyError, DEV_ROLE_MAP.slug, "_rubbish_")
    assert_raises(KeyError, DEV_ROLE_MAP.slug, None)

    assert (
        "HTG" not in DEV_ROLE_MAP
        and "0E" in DEV_ROLE_MAP
        and "heating_relay" not in DEV_ROLE_MAP
    )
    assert (
        "DHW" not in DEV_ROLE_MAP.values()
        and "0D" not in DEV_ROLE_MAP.values()
        and "dhw_sensor" in DEV_ROLE_MAP.values()
    )

    assert (
        "DHW" in DEV_ROLE_MAP.slugs()
        and "0D" not in DEV_ROLE_MAP.slugs()
        and "dhw_sensor" not in DEV_ROLE_MAP.slugs()
    )

    assert DEV_ROLE_MAP.SLUGS, MAIN_SLUGS
    assert DEV_ROLE_MAP.HEAT_DEVICES == ("00", "04", "08", "09", "0A", "0B", "11")


MAIN_DICT = {
    "ACT": {"00": "zone_actuator"},
    "SEN": {"04": "zone_sensor"},
    "RAD": {"08": "rad_actuator"},
    "UFH": {"09": "ufh_actuator"},
    "VAL": {"0A": "val_actuator"},
    "MIX": {"0B": "mix_actuator"},
    "OUT": {"0C": "out_sensor"},
    "DHW": {"0D": "dhw_sensor"},
    "HTG": {"0E": "hotwater_valve"},
    "HT1": {None: "heating_valve"},
    "APP": {"0F": "appliance_control"},
    "RFG": {"10": "remote_gateway"},
    "ELE": {"11": "ele_actuator"},
}
ATTR_DICT = {
    "HEAT_DEVICES": ("08", "09", "0A", "0B", "11"),
}
MAIN_SLUGS = tuple(sorted(MAIN_DICT.keys()))
